# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.test import APITestCase
from unittest.mock import patch

from swh.storage.exc import StorageDBError, StorageAPIError

from swh.web.tests.testcase import WebTestCase


class OriginApiTestCase(WebTestCase, APITestCase):

    def setUp(self):
        self.origin_visit1 = {
            'date': 1104616800.0,
            'origin': 10,
            'visit': 100,
            'metadata': None,
            'status': 'full',
        }

        self.origin1 = {
            'id': 1234,
            'url': 'ftp://some/url/to/origin/0',
            'type': 'ftp'
        }

    @patch('swh.web.api.views.origin.get_origin_visits')
    def test_api_1_lookup_origin_visits_raise_error(
        self, mock_get_origin_visits,
    ):
        # given
        mock_get_origin_visits.side_effect = ValueError(
            'voluntary error to check the bad request middleware.')
        # when
        rv = self.client.get('/api/1/origin/2/visits/')
        # then
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'ValueError',
            'reason': 'voluntary error to check the bad request middleware.'})

    @patch('swh.web.api.views.origin.get_origin_visits')
    def test_api_1_lookup_origin_visits_raise_swh_storage_error_db(
            self, mock_get_origin_visits):
        # given
        mock_get_origin_visits.side_effect = StorageDBError(
            'Storage exploded! Will be back online shortly!')
        # when
        rv = self.client.get('/api/1/origin/2/visits/')
        # then
        self.assertEqual(rv.status_code, 503)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'StorageDBError',
            'reason':
            'An unexpected error occurred in the backend: '
            'Storage exploded! Will be back online shortly!'})

    @patch('swh.web.api.views.origin.get_origin_visits')
    def test_api_1_lookup_origin_visits_raise_swh_storage_error_api(
            self, mock_get_origin_visits):
        # given
        mock_get_origin_visits.side_effect = StorageAPIError(
            'Storage API dropped dead! Will resurrect from its ashes asap!'
        )
        # when
        rv = self.client.get('/api/1/origin/2/visits/')
        # then
        self.assertEqual(rv.status_code, 503)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'StorageAPIError',
            'reason':
            'An unexpected error occurred in the api backend: '
            'Storage API dropped dead! Will resurrect from its ashes asap!'
        })

    @patch('swh.web.api.views.origin.get_origin_visits')
    def test_api_1_lookup_origin_visits(self, mock_get_origin_visits):
        # given
        stub_visits = [
            {
                'date': 1293919200.0,
                'origin': 2,
                'snapshot': '1234',
                'visit': 1
            },
            {
                'date': 1293919200.0,
                'origin': 2,
                'snapshot': '1234',
                'visit': 2
            },
            {
                'date': 1420149600.0,
                'origin': 2,
                'snapshot': '5678',
                'visit': 3
            },
            {
                'date': 1420149600.0,
                'origin': 2,
                'snapshot': '5678',
                'visit': 4
            }
        ]

        mock_get_origin_visits.return_value = stub_visits

        # when
        rv = self.client.get('/api/1/origin/2/visits/?per_page=2&last_visit=3')

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, [
            {
                'date': 1293919200.0,
                'origin': 2,
                'snapshot': '1234',
                'visit': 2,
                'origin_visit_url': '/api/1/origin/2/visit/2/',
                'snapshot_url': '/api/1/snapshot/1234/'
            },
            {
                'date': 1293919200.0,
                'origin': 2,
                'snapshot': '1234',
                'visit': 1,
                'origin_visit_url': '/api/1/origin/2/visit/1/',
                'snapshot_url': '/api/1/snapshot/1234/'
            },

        ])

    @patch('swh.web.api.views.origin.service')
    def test_api_1_lookup_origin_visit(self, mock_service):
        # given
        origin_visit = self.origin_visit1.copy()
        origin_visit.update({
            'snapshot': '57478754'
        })

        mock_service.lookup_origin_visit.return_value = origin_visit

        expected_origin_visit = self.origin_visit1.copy()
        expected_origin_visit.update({
            'origin_url': '/api/1/origin/10/',
            'snapshot': '57478754',
            'snapshot_url': '/api/1/snapshot/57478754/'
        })

        # when
        rv = self.client.get('/api/1/origin/10/visit/100/')

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, expected_origin_visit)

        mock_service.lookup_origin_visit.assert_called_once_with('10', '100')

    @patch('swh.web.api.views.origin.service')
    def test_api_1_lookup_origin_visit_not_found(self, mock_service):
        # given
        mock_service.lookup_origin_visit.return_value = None

        # when
        rv = self.client.get('/api/1/origin/1/visit/1000/')

        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'No visit 1000 for origin 1 found'
        })

        mock_service.lookup_origin_visit.assert_called_once_with('1', '1000')

    @patch('swh.web.api.views.origin.service')
    def test_api_origin_by_id(self, mock_service):
        # given
        mock_service.lookup_origin.return_value = self.origin1

        expected_origin = self.origin1.copy()
        expected_origin.update({
            'origin_visits_url': '/api/1/origin/1234/visits/'
        })

        # when
        rv = self.client.get('/api/1/origin/1234/')

        # then
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, expected_origin)

        mock_service.lookup_origin.assert_called_with({'id': '1234'})

    @patch('swh.web.api.views.origin.service')
    def test_api_origin_by_type_url(self, mock_service):
        # given
        stub_origin = self.origin1.copy()
        stub_origin.update({
            'id': 987
        })
        mock_service.lookup_origin.return_value = stub_origin

        expected_origin = stub_origin.copy()
        expected_origin.update({
            'origin_visits_url': '/api/1/origin/987/visits/'
        })

        # when
        rv = self.client.get('/api/1/origin/ftp/url'
                             '/ftp://some/url/to/origin/0/')

        # then
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, expected_origin)

        mock_service.lookup_origin.assert_called_with(
            {'url': 'ftp://some/url/to/origin/0',
             'type': 'ftp'})

    @patch('swh.web.api.views.origin.service')
    def test_api_origin_not_found(self, mock_service):
        # given
        mock_service.lookup_origin.return_value = None

        # when
        rv = self.client.get('/api/1/origin/4321/')

        # then
        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'Origin with id 4321 not found.'
        })

        mock_service.lookup_origin.assert_called_with({'id': '4321'})

    @patch('swh.web.common.service.idx_storage')
    def test_api_origin_metadata_search(self, mock_idx_storage):
        # given
        mock_idx_storage.origin_intrinsic_metadata_search_fulltext \
            .return_value = [{
                'from_revision':
                b'p&\xb7\xc1\xa2\xafVR\x1e\x95\x1c\x01\xed \xf2U\xfa\x05B8',
                'metadata': {'author': 'Jane Doe'},
                'origin_id': 54974445,
                'tool': {
                    'configuration': {
                        'context': ['NpmMapping', 'CodemetaMapping'],
                        'type': 'local'
                    },
                    'id': 3,
                    'name': 'swh-metadata-detector',
                    'version': '0.0.1'
                }
            }]

        # when
        rv = self.client.get(
            '/api/1/origin/metadata-search/?fulltext=Jane%20Doe')

        # then
        self.assertEqual(rv.status_code, 200, rv.content)
        self.assertEqual(rv['Content-Type'], 'application/json')
        expected_data = [{
            'origin_id': 54974445,
            'metadata': {'author': 'Jane Doe'},
            'from_revision': '7026b7c1a2af56521e951c01ed20f255fa054238',
            'tool': {
                'configuration': {
                    'context': ['NpmMapping', 'CodemetaMapping'],
                    'type': 'local'
                },
                'id': 3,
                'name': 'swh-metadata-detector',
                'version': '0.0.1',
            }
        }]
        self.assertEqual(rv.data, expected_data)
        mock_idx_storage.origin_intrinsic_metadata_search_fulltext \
            .assert_called_with(conjunction=['Jane Doe'], limit=70)

    @patch('swh.web.common.service.idx_storage')
    def test_api_origin_metadata_search_limit(self, mock_idx_storage):
        # given
        mock_idx_storage.origin_intrinsic_metadata_search_fulltext \
            .return_value = [{
                'from_revision':
                b'p&\xb7\xc1\xa2\xafVR\x1e\x95\x1c\x01\xed \xf2U\xfa\x05B8',
                'metadata': {'author': 'Jane Doe'},
                'origin_id': 54974445,
                'tool': {
                    'configuration': {
                        'context': ['NpmMapping', 'CodemetaMapping'],
                        'type': 'local'
                    },
                    'id': 3,
                    'name': 'swh-metadata-detector',
                    'version': '0.0.1'
                }
            }]

        # when
        rv = self.client.get(
            '/api/1/origin/metadata-search/?fulltext=Jane%20Doe')

        # then
        self.assertEqual(rv.status_code, 200, rv.content)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(len(rv.data), 1)
        mock_idx_storage.origin_intrinsic_metadata_search_fulltext \
            .assert_called_with(conjunction=['Jane Doe'], limit=70)

        # when
        rv = self.client.get(
            '/api/1/origin/metadata-search/?fulltext=Jane%20Doe&limit=10')

        # then
        self.assertEqual(rv.status_code, 200, rv.content)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(len(rv.data), 1)
        mock_idx_storage.origin_intrinsic_metadata_search_fulltext \
            .assert_called_with(conjunction=['Jane Doe'], limit=10)

        # when
        rv = self.client.get(
            '/api/1/origin/metadata-search/?fulltext=Jane%20Doe&limit=987')

        # then
        self.assertEqual(rv.status_code, 200, rv.content)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(len(rv.data), 1)
        mock_idx_storage.origin_intrinsic_metadata_search_fulltext \
            .assert_called_with(conjunction=['Jane Doe'], limit=100)

    @patch('swh.web.common.service.idx_storage')
    def test_api_origin_metadata_search_invalid(self, mock_idx_storage):
        rv = self.client.get('/api/1/origin/metadata-search/')

        # then
        self.assertEqual(rv.status_code, 400, rv.content)
        mock_idx_storage.assert_not_called()



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import shutil
from subprocess import run, PIPE

from django.core.cache import cache
from hypothesis.extra.django import TestCase

from swh.model.hashutil import hash_to_bytes
from swh.web import config
from swh.web.common import converters, service
from swh.web.tests.data import get_tests_data


ctags_json_missing = \
    shutil.which('ctags') is None or \
    b'+json' not in run(['ctags', '--version'], stdout=PIPE).stdout

fossology_missing = shutil.which('nomossa') is None


class WebTestCase(TestCase):
    """Base TestCase class for swh-web.

    It is initialized with references to in-memory storages containing
    raw tests data.

    It also defines class methods to retrieve those tests data in
    a json serializable format in order to ease tests implementation.

    """
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        tests_data = get_tests_data()
        cls.storage = tests_data['storage']
        cls.idx_storage = tests_data['idx_storage']
        cls.mimetype_indexer = tests_data['mimetype_indexer']
        cls.language_indexer = tests_data['language_indexer']
        cls.license_indexer = tests_data['license_indexer']
        cls.ctags_indexer = tests_data['ctags_indexer']

        # Update swh-web configuration to use the in-memory storage
        # instantiated in the tests.data module
        swh_config = config.get_config()
        swh_config.update({'storage': cls.storage})
        service.storage = cls.storage

        # Update swh-web configuration to use the in-memory indexer storage
        # instantiated in the tests.data modules
        swh_config.update({'indexer_storage': cls.idx_storage})
        service.idx_storage = cls.idx_storage

    @classmethod
    def content_add_mimetype(cls, cnt_id):
        cls.mimetype_indexer.run([hash_to_bytes(cnt_id)],
                                 'update-dups')

    @classmethod
    def content_get_mimetype(cls, cnt_id):
        mimetype = next(cls.idx_storage.content_mimetype_get(
                        [hash_to_bytes(cnt_id)]))
        return converters.from_filetype(mimetype)

    @classmethod
    def content_add_language(cls, cnt_id):
        cls.language_indexer.run([hash_to_bytes(cnt_id)],
                                 'update-dups')

    @classmethod
    def content_get_language(cls, cnt_id):
        lang = next(cls.idx_storage.content_language_get(
                    [hash_to_bytes(cnt_id)]))
        return converters.from_swh(lang, hashess={'id'})

    @classmethod
    def content_add_license(cls, cnt_id):
        cls.license_indexer.run([hash_to_bytes(cnt_id)],
                                'update-dups')

    @classmethod
    def content_get_license(cls, cnt_id):
        cnt_id_bytes = hash_to_bytes(cnt_id)
        lic = next(cls.idx_storage.content_fossology_license_get(
                   [cnt_id_bytes]))
        return converters.from_swh({'id': cnt_id_bytes,
                                    'facts': lic[cnt_id_bytes]},
                                   hashess={'id'})

    @classmethod
    def content_add_ctags(cls, cnt_id):
        cls.ctags_indexer.run([hash_to_bytes(cnt_id)],
                              'update-dups')

    @classmethod
    def content_get_ctags(cls, cnt_id):
        cnt_id_bytes = hash_to_bytes(cnt_id)
        ctags = cls.idx_storage.content_ctags_get([cnt_id_bytes])
        for ctag in ctags:
            yield converters.from_swh(ctag, hashess={'id'})

    @classmethod
    def content_get_metadata(cls, cnt_id):
        cnt_id_bytes = hash_to_bytes(cnt_id)
        metadata = next(cls.storage.content_get_metadata([cnt_id_bytes]))
        return converters.from_swh(metadata,
                                   hashess={'sha1', 'sha1_git', 'sha256',
                                            'blake2s256'})

    @classmethod
    def content_get(cls, cnt_id):
        cnt_id_bytes = hash_to_bytes(cnt_id)
        cnt = next(cls.storage.content_get([cnt_id_bytes]))
        return converters.from_content(cnt)

    @classmethod
    def directory_ls(cls, dir_id):
        cnt_id_bytes = hash_to_bytes(dir_id)
        dir_content = map(converters.from_directory_entry,
                          cls.storage.directory_ls(cnt_id_bytes))
        return list(dir_content)

    @classmethod
    def release_get(cls, rel_id):
        rel_id_bytes = hash_to_bytes(rel_id)
        rel_data = next(cls.storage.release_get([rel_id_bytes]))
        return converters.from_release(rel_data)

    @classmethod
    def revision_get(cls, rev_id):
        rev_id_bytes = hash_to_bytes(rev_id)
        rev_data = next(cls.storage.revision_get([rev_id_bytes]))
        return converters.from_revision(rev_data)

    @classmethod
    def revision_log(cls, rev_id):
        rev_id_bytes = hash_to_bytes(rev_id)
        return map(converters.from_revision,
                   cls.storage.revision_log([rev_id_bytes]))

    def setUp(self):
        cache.clear()



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.test import APITestCase
from unittest.mock import patch

from swh.model import hashutil

from swh.web.tests.testcase import WebTestCase

TEST_OBJ_ID = 'd4905454cc154b492bd6afed48694ae3c579345e'

OBJECT_TYPES = {'directory': ('directory', None),
                'revision_gitfast': ('revision', 'gitfast')}


class VaultApiTestCase(WebTestCase, APITestCase):
    @patch('swh.web.api.views.vault.service')
    def test_api_vault_cook(self, mock_service):
        stub_cook = {
            'fetch_url': ('http://127.0.0.1:5004/api/1/vault/directory/{}/raw/'
                          .format(TEST_OBJ_ID)),
            'obj_id': 'd4905454cc154b492bd6afed48694ae3c579345e',
            'obj_type': 'test_type',
            'progress_message': None,
            'status': 'done',
            'task_uuid': 'de75c902-5ee5-4739-996e-448376a93eff',
        }
        stub_fetch = b'content'

        mock_service.vault_cook.return_value = stub_cook
        mock_service.vault_fetch.return_value = stub_fetch

        for obj_type, (obj_type_name, obj_type_format) in OBJECT_TYPES.items():
            url = '/api/1/vault/{}/{}/'.format(obj_type_name, TEST_OBJ_ID)
            if obj_type_format:
                url += '{}/'.format(obj_type_format)
            rv = self.client.post(url, {'email': 'test@test.mail'})

            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv['Content-Type'], 'application/json')

            self.assertEqual(rv.data, stub_cook)
            mock_service.vault_cook.assert_called_with(
                obj_type,
                hashutil.hash_to_bytes(TEST_OBJ_ID),
                'test@test.mail')

            rv = self.client.get(url + 'raw/')

            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv['Content-Type'], 'application/gzip')
            self.assertEqual(rv.content, stub_fetch)
            mock_service.vault_fetch.assert_called_with(
                obj_type, hashutil.hash_to_bytes(TEST_OBJ_ID))

    @patch('swh.web.api.views.vault.service')
    def test_api_vault_cook_notfound(self, mock_service):
        mock_service.vault_cook.return_value = None
        mock_service.vault_fetch.return_value = None

        for obj_type, (obj_type_name, obj_type_format) in OBJECT_TYPES.items():
            url = '/api/1/vault/{}/{}/'.format(obj_type_name, TEST_OBJ_ID)
            if obj_type_format:
                url += '{}/'.format(obj_type_format)
            rv = self.client.post(url)

            self.assertEqual(rv.status_code, 404)
            self.assertEqual(rv['Content-Type'], 'application/json')

            self.assertEqual(rv.data['exception'], 'NotFoundExc')
            mock_service.vault_cook.assert_called_with(
                obj_type, hashutil.hash_to_bytes(TEST_OBJ_ID), None)

            rv = self.client.get(url + 'raw/')

            self.assertEqual(rv.status_code, 404)
            self.assertEqual(rv['Content-Type'], 'application/json')
            self.assertEqual(rv.data['exception'], 'NotFoundExc')
            mock_service.vault_fetch.assert_called_with(
                obj_type, hashutil.hash_to_bytes(TEST_OBJ_ID))



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.test import APITestCase
from rest_framework.response import Response

from swh.web.api.apidoc import api_doc, _parse_httpdomain_doc
from swh.web.api.apiurls import api_route
from swh.web.tests.testcase import WebTestCase

# flake8: noqa

httpdomain_doc = """
.. http:get:: /api/1/revision/(sha1_git)/

    Get information about a revision in the archive.
    Revisions are identified by **sha1** checksums, compatible with Git commit identifiers.
    See :func:`swh.model.identifiers.revision_identifier` in our data model module for details
    about how they are computed.

    :param string sha1_git: hexadecimal representation of the revision **sha1_git** identifier

    :reqheader Accept: the requested response content type,
        either ``application/json`` (default) or ``application/yaml``
    :resheader Content-Type: this depends on :http:header:`Accept` header of request

    :>json object author: information about the author of the revision
    :>json string author_url: link to :http:get:`/api/1/person/(person_id)/` to get
        information about the author of the revision
    :>json object committer: information about the committer of the revision
    :>json string committer_url: link to :http:get:`/api/1/person/(person_id)/` to get
        information about the committer of the revision
    :>json string committer_date: ISO representation of the commit date (in UTC)
    :>json string date: ISO representation of the revision date (in UTC)
    :>json string directory: the unique identifier that revision points to
    :>json string directory_url: link to :http:get:`/api/1/directory/(sha1_git)/[(path)/]`
        to get information about the directory associated to the revision
    :>json string id: the revision unique identifier
    :>json boolean merge: whether or not the revision corresponds to a merge commit
    :>json string message: the message associated to the revision
    :>json array parents: the parents of the revision, i.e. the previous revisions
        that head directly to it, each entry of that array contains an unique parent
        revision identifier but also a link to :http:get:`/api/1/revision/(sha1_git)/`
        to get more information about it
    :>json string type: the type of the revision

    **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

    :statuscode 200: no error
    :statuscode 400: an invalid **sha1_git** value has been provided
    :statuscode 404: requested revision can not be found in the archive

    **Request:**

    .. parsed-literal::

        $ curl -i :swh_web_api:`revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/`
"""


class APIDocTestCase(WebTestCase, APITestCase):

    def test_apidoc_nodoc_failure(self):
        with self.assertRaises(Exception):
            @api_doc('/my/nodoc/url/')
            def apidoc_nodoc_tester(request, arga=0, argb=0):
                return Response(arga + argb)

    @staticmethod
    @api_route(r'/some/(?P<myarg>[0-9]+)/(?P<myotherarg>[0-9]+)/',
               'some-doc-route')
    @api_doc('/some/doc/route/')
    def apidoc_route(request, myarg, myotherarg, akw=0):
        """
        Sample doc
        """
        return {'result': int(myarg) + int(myotherarg) + akw}

    def test_apidoc_route_doc(self):
        # when
        rv = self.client.get('/api/1/some/doc/route/')

        # then
        self.assertEqual(rv.status_code, 200)
        self.assertTemplateUsed('api/apidoc.html')

    def test_apidoc_route_fn(self):

        # when
        rv = self.client.get('/api/1/some/1/1/')

        # then
        self.assertEqual(rv.status_code, 200)

    @staticmethod
    @api_route(r'/some/full/(?P<myarg>[0-9]+)/(?P<myotherarg>[0-9]+)/',
               'some-complete-doc-route')
    @api_doc('/some/complete/doc/route/')
    def apidoc_full_stack(request, myarg, myotherarg, akw=0):
        """
        Sample doc
        """
        return {'result': int(myarg) + int(myotherarg) + akw}

    def test_apidoc_full_stack_doc(self):
        # when
        rv = self.client.get('/api/1/some/complete/doc/route/')

        # then
        self.assertEqual(rv.status_code, 200)
        self.assertTemplateUsed('api/apidoc.html')

    def test_apidoc_full_stack_fn(self):
        # when
        rv = self.client.get('/api/1/some/full/1/1/')

        # then
        self.assertEqual(rv.status_code, 200)

    def test_api_doc_parse_httpdomain(self):
        doc_data = {
            'description': '',
            'urls': [],
            'args': [],
            'params': [],
            'resheaders': [],
            'reqheaders': [],
            'return_type': '',
            'returns': [],
            'status_codes': [],
            'examples': []
        }

        _parse_httpdomain_doc(httpdomain_doc, doc_data)

        expected_urls = [{
            'rule': '/api/1/revision/ **\\(sha1_git\\)** /',
            'methods': ['GET', 'HEAD', 'OPTIONS']
        }]

        self.assertIn('urls', doc_data)
        self.assertEqual(doc_data['urls'], expected_urls)

        expected_description = 'Get information about a revision in the archive. \
Revisions are identified by **sha1** checksums, compatible with Git commit \
identifiers. See **swh.model.identifiers.revision_identifier** in our data \
model module for details about how they are computed.'
        self.assertIn('description', doc_data)
        self.assertEqual(doc_data['description'], expected_description)

        expected_args = [{
            'name': 'sha1_git',
            'type': 'string',
            'doc': 'hexadecimal representation of the revision **sha1_git** identifier'
        }]

        self.assertIn('args', doc_data)
        self.assertEqual(doc_data['args'], expected_args)

        expected_params = []
        self.assertIn('params', doc_data)
        self.assertEqual(doc_data['params'], expected_params)

        expected_reqheaders = [{
            'doc': 'the requested response content type, either ``application/json``  or ``application/yaml``',
            'name': 'Accept'
        }]

        self.assertIn('reqheaders', doc_data)
        self.assertEqual(doc_data['reqheaders'], expected_reqheaders)

        expected_resheaders = [{
            'doc': 'this depends on **Accept** header of request',
            'name': 'Content-Type'
        }]

        self.assertIn('resheaders', doc_data)
        self.assertEqual(doc_data['resheaders'], expected_resheaders)

        expected_statuscodes = [
            {
                'code': '200',
                'doc': 'no error'
            },
            {
                'code': '400',
                'doc': 'an invalid **sha1_git** value has been provided'
            },
            {
                'code': '404',
                'doc': 'requested revision can not be found in the archive'
            }
        ]

        self.assertIn('status_codes', doc_data)
        self.assertEqual(doc_data['status_codes'], expected_statuscodes)

        expected_return_type = 'object'

        self.assertIn('return_type', doc_data)
        self.assertEqual(doc_data['return_type'], expected_return_type)

        expected_returns = [
            {
                'name': 'author',
                'type': 'object',
                'doc': 'information about the author of the revision'
            },
            {
                'name': 'author_url',
                'type': 'string',
                'doc': 'link to `</api/1/person/>`_ to get information about the author of the revision'
            },
            {
                'name': 'committer',
                'type': 'object',
                'doc': 'information about the committer of the revision'
            },
            {
                'name': 'committer_url',
                'type': 'string',
                'doc': 'link to `</api/1/person/>`_ to get information about the committer of the revision'
            },
            {
                'name': 'committer_date',
                'type': 'string',
                'doc': 'ISO representation of the commit date (in UTC)'
            },
            {
                'name': 'date',
                'type': 'string',
                'doc': 'ISO representation of the revision date (in UTC)'
            },
            {
                'name': 'directory',
                'type': 'string',
                'doc': 'the unique identifier that revision points to'
            },
            {
                'name': 'directory_url',
                'type': 'string',
                'doc': 'link to `</api/1/directory/>`_ to get information about the directory associated to the revision'
            },
            {
                'name': 'id',
                'type': 'string',
                'doc': 'the revision unique identifier'
            },
            {
                'name': 'merge',
                'type': 'boolean',
                'doc': 'whether or not the revision corresponds to a merge commit'
            },
            {
                'name': 'message',
                'type': 'string',
                'doc': 'the message associated to the revision'
            },
            {
                'name': 'parents',
                'type': 'array',
                'doc': 'the parents of the revision, i.e. the previous revisions that head directly to it, each entry of that array contains an unique parent revision identifier but also a link to `</api/1/revision/>`_ to get more information about it'
            },
            {
                'name': 'type',
                'type': 'string',
                'doc': 'the type of the revision'
            }
        ]

        self.assertIn('returns', doc_data)
        self.assertEqual(doc_data['returns'], expected_returns)

        expected_examples = ['/api/1/revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/']

        self.assertIn('examples', doc_data)
        self.assertEqual(doc_data['examples'], expected_examples)



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.common.urlsindex import UrlsIndex


class BrowseUrls(UrlsIndex):
    """
    Class to manage swh-web browse application urls.
    """

    scope = 'browse'


class browse_route(object):  # noqa: N801
    """
    Decorator to ease the registration of a swh-web browse endpoint

    Args:
        url_patterns: list of url patterns used by Django to identify the browse routes
        view_name: the name of the Django view associated to the routes used to
           reverse the url
    """ # noqa

    def __init__(self, *url_patterns, view_name=None):
        super().__init__()
        self.url_patterns = []
        for url_pattern in url_patterns:
            self.url_patterns.append('^' + url_pattern + '$')
        self.view_name = view_name

    def __call__(self, f):
        # register the route and its view in the browse endpoints index
        for url_pattern in self.url_patterns:
            BrowseUrls.add_url_pattern(url_pattern, f, self.view_name)
        return f



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from rest_framework.test import APIRequestFactory

from unittest.mock import patch

from swh.web.api.apiresponse import (
    compute_link_header, transform, make_api_response,
    filter_by_fields
)
from swh.web.tests.testcase import WebTestCase

api_request_factory = APIRequestFactory()


class SWHComputeLinkHeaderTest(WebTestCase):
    def test_compute_link_header(self):
        rv = {
            'headers': {'link-next': 'foo', 'link-prev': 'bar'},
            'results': [1, 2, 3]
        }
        options = {}

        # when
        headers = compute_link_header(
            rv, options)

        self.assertEqual(headers, {
            'Link': '<foo>; rel="next",<bar>; rel="previous"',
        })

    def test_compute_link_header_nothing_changed(self):
        rv = {}
        options = {}

        # when
        headers = compute_link_header(
            rv, options)

        self.assertEqual(headers, {})

    def test_compute_link_header_nothing_changed_2(self):
        rv = {'headers': {}}
        options = {}

        # when
        headers = compute_link_header(
            rv, options)

        self.assertEqual(headers, {})


class SWHTransformProcessorTest(WebTestCase):
    def test_transform_only_return_results_1(self):
        rv = {'results': {'some-key': 'some-value'}}

        self.assertEqual(transform(rv), {'some-key': 'some-value'})

    def test_transform_only_return_results_2(self):
        rv = {'headers': {'something': 'do changes'},
              'results': {'some-key': 'some-value'}}

        self.assertEqual(transform(rv), {'some-key': 'some-value'})

    def test_transform_do_remove_headers(self):
        rv = {'headers': {'something': 'do changes'},
              'some-key': 'some-value'}

        self.assertEqual(transform(rv), {'some-key': 'some-value'})

    def test_transform_do_nothing(self):
        rv = {'some-key': 'some-value'}

        self.assertEqual(transform(rv), {'some-key': 'some-value'})


class RendererTestCase(WebTestCase):

    @patch('swh.web.api.apiresponse.json')
    @patch('swh.web.api.apiresponse.filter_by_fields')
    @patch('swh.web.api.apiresponse.shorten_path')
    def test_swh_multi_response_mimetype(self, mock_shorten_path,
                                         mock_filter, mock_json):
        # given
        data = {
            'data': [12, 34],
            'id': 'adc83b19e793491b1c6ea0fd8b46cd9f32e592fc'
        }

        mock_filter.return_value = data
        mock_shorten_path.return_value = 'my_short_path'

        accepted_response_formats = {'html': 'text/html',
                                     'yaml': 'application/yaml',
                                     'json': 'application/json'}

        for format in accepted_response_formats:

            request = api_request_factory.get('/api/test/path/')

            mime_type = accepted_response_formats[format]
            setattr(request, 'accepted_media_type', mime_type)

            if mime_type == 'text/html':

                expected_data = {
                    'response_data': json.dumps(data),
                    'request': {
                        'path': request.path,
                        'method': request.method,
                        'absolute_uri': request.build_absolute_uri()
                    },
                    'headers_data': {},
                    'heading': 'my_short_path',
                    'status_code': 200
                }

                mock_json.dumps.return_value = json.dumps(data)
            else:
                expected_data = data

            # when

            rv = make_api_response(request, data)

            # then
            mock_filter.assert_called_with(request, data)
            self.assertEqual(rv.data, expected_data)
            self.assertEqual(rv.status_code, 200)
            if mime_type == 'text/html':
                self.assertEqual(rv.template_name, 'api/apidoc.html')

    def test_swh_filter_renderer_do_nothing(self):
        # given
        input_data = {'a': 'some-data'}

        request = api_request_factory.get('/api/test/path/', data={})
        setattr(request, 'query_params', request.GET)

        # when
        actual_data = filter_by_fields(request, input_data)

        # then
        self.assertEqual(actual_data, input_data)

    @patch('swh.web.api.apiresponse.utils.filter_field_keys')
    def test_swh_filter_renderer_do_filter(self, mock_ffk):
        # given
        mock_ffk.return_value = {'a': 'some-data'}

        request = api_request_factory.get('/api/test/path/',
                                          data={'fields': 'a,c'})
        setattr(request, 'query_params', request.GET)

        input_data = {'a': 'some-data',
                      'b': 'some-other-data'}

        # when
        actual_data = filter_by_fields(request, input_data)

        # then
        self.assertEqual(actual_data, {'a': 'some-data'})

        mock_ffk.assert_called_once_with(input_data, {'a', 'c'})



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.test import APITestCase
from unittest.mock import patch

from swh.storage.exc import StorageDBError, StorageAPIError

from swh.web.tests.testcase import WebTestCase


class StatApiTestCase(WebTestCase, APITestCase):
    @patch('swh.web.api.views.stat.service')
    def test_api_1_stat_counters_raise_error(self, mock_service):
        # given
        mock_service.stat_counters.side_effect = ValueError(
            'voluntary error to check the bad request middleware.')
        # when
        rv = self.client.get('/api/1/stat/counters/')
        # then
        self.assertEqual(rv.status_code, 400)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'ValueError',
            'reason': 'voluntary error to check the bad request middleware.'})

    @patch('swh.web.api.views.stat.service')
    def test_api_1_stat_counters_raise_from_db(self, mock_service):
        # given
        mock_service.stat_counters.side_effect = StorageDBError(
            'Storage exploded! Will be back online shortly!')
        # when
        rv = self.client.get('/api/1/stat/counters/')
        # then
        self.assertEqual(rv.status_code, 503)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'StorageDBError',
            'reason':
            'An unexpected error occurred in the backend: '
            'Storage exploded! Will be back online shortly!'})

    @patch('swh.web.api.views.stat.service')
    def test_api_1_stat_counters_raise_from_api(self, mock_service):
        # given
        mock_service.stat_counters.side_effect = StorageAPIError(
            'Storage API dropped dead! Will resurrect from its ashes asap!'
        )
        # when
        rv = self.client.get('/api/1/stat/counters/')
        # then
        self.assertEqual(rv.status_code, 503)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'StorageAPIError',
            'reason':
            'An unexpected error occurred in the api backend: '
            'Storage API dropped dead! Will resurrect from its ashes asap!'
        })

    @patch('swh.web.api.views.stat.service')
    def test_api_1_stat_counters(self, mock_service):
        # given
        stub_stats = {
            "content": 1770830,
            "directory": 211683,
            "directory_entry_dir": 209167,
            "directory_entry_file": 1807094,
            "directory_entry_rev": 0,
            "origin": 1096,
            "person": 0,
            "release": 8584,
            "revision": 7792,
            "revision_history": 0,
            "skipped_content": 0
        }
        mock_service.stat_counters.return_value = stub_stats

        # when
        rv = self.client.get('/api/1/stat/counters/')

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, stub_stats)

        mock_service.stat_counters.assert_called_once_with()



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.web.common import highlightjs
from swh.web.tests.testcase import WebTestCase


class HighlightJsTestCase(WebTestCase):

    def test_get_hljs_language_from_mime_type(self):

        lang = highlightjs.get_hljs_language_from_mime_type('text/plain')
        self.assertEqual(lang, None)

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-c')
        self.assertEqual(lang, 'cpp')

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-c++')
        self.assertEqual(lang, 'cpp')

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-perl')
        self.assertEqual(lang, 'perl')

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-python')
        self.assertEqual(lang, 'python')

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-msdos-batch') # noqa
        self.assertEqual(lang, 'dos')

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-tex')
        self.assertEqual(lang, 'tex')

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-lisp')
        self.assertEqual(lang, 'lisp')

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-java')
        self.assertEqual(lang, 'java')

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-makefile')
        self.assertEqual(lang, 'makefile')

        lang = highlightjs.get_hljs_language_from_mime_type('text/x-shellscript') # noqa
        self.assertEqual(lang, 'bash')

        lang = highlightjs.get_hljs_language_from_mime_type('image/png')
        self.assertEqual(lang, None)

    def test_get_hljs_language_from_filename(self):

        lang = highlightjs.get_hljs_language_from_filename('foo')
        self.assertEqual(lang, None)

        lang = highlightjs.get_hljs_language_from_filename('foo.h')
        self.assertEqual(lang, 'cpp')

        lang = highlightjs.get_hljs_language_from_filename('foo.c')
        self.assertEqual(lang, 'cpp')

        lang = highlightjs.get_hljs_language_from_filename('foo.c.in')
        self.assertEqual(lang, 'cpp')

        lang = highlightjs.get_hljs_language_from_filename('foo.cpp')
        self.assertEqual(lang, 'cpp')

        lang = highlightjs.get_hljs_language_from_filename('foo.pl')
        self.assertEqual(lang, 'perl')

        lang = highlightjs.get_hljs_language_from_filename('foo.py')
        self.assertEqual(lang, 'python')

        lang = highlightjs.get_hljs_language_from_filename('foo.md')
        self.assertEqual(lang, 'markdown')

        lang = highlightjs.get_hljs_language_from_filename('foo.js')
        self.assertEqual(lang, 'javascript')

        lang = highlightjs.get_hljs_language_from_filename('foo.bat')
        self.assertEqual(lang, 'dos')

        lang = highlightjs.get_hljs_language_from_filename('foo.json')
        self.assertEqual(lang, 'json')

        lang = highlightjs.get_hljs_language_from_filename('foo.yml')
        self.assertEqual(lang, 'yaml')

        lang = highlightjs.get_hljs_language_from_filename('foo.ini')
        self.assertEqual(lang, 'ini')

        lang = highlightjs.get_hljs_language_from_filename('foo.cfg')
        self.assertEqual(lang, 'ini')

        lang = highlightjs.get_hljs_language_from_filename('foo.hy')
        self.assertEqual(lang, 'hy')

        lang = highlightjs.get_hljs_language_from_filename('foo.lisp')
        self.assertEqual(lang, 'lisp')

        lang = highlightjs.get_hljs_language_from_filename('foo.java')
        self.assertEqual(lang, 'java')

        lang = highlightjs.get_hljs_language_from_filename('foo.sh')
        self.assertEqual(lang, 'bash')

        lang = highlightjs.get_hljs_language_from_filename('foo.cmake')
        self.assertEqual(lang, 'cmake')

        lang = highlightjs.get_hljs_language_from_filename('foo.ml')
        self.assertEqual(lang, 'ocaml')

        lang = highlightjs.get_hljs_language_from_filename('foo.mli')
        self.assertEqual(lang, 'ocaml')

        lang = highlightjs.get_hljs_language_from_filename('foo.rb')
        self.assertEqual(lang, 'ruby')

        lang = highlightjs.get_hljs_language_from_filename('foo.jl')
        self.assertEqual(lang, 'julia')

        lang = highlightjs.get_hljs_language_from_filename('Makefile')
        self.assertEqual(lang, 'makefile')

        lang = highlightjs.get_hljs_language_from_filename('CMakeLists.txt')
        self.assertEqual(lang, 'cmake')



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from hypothesis import given
from rest_framework.test import APITestCase

from swh.web.common.utils import reverse
from swh.web.tests.strategies import (
    content, unknown_content, contents_with_ctags
)
from swh.web.tests.testcase import (
    WebTestCase, ctags_json_missing, fossology_missing
)


class ContentApiTestCase(WebTestCase, APITestCase):

    @given(content())
    def test_api_content_filetype(self, content):

        self.content_add_mimetype(content['sha1'])
        url = reverse('api-content-filetype',
                      url_args={'q': 'sha1_git:%s' % content['sha1_git']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        content_url = reverse('api-content',
                              url_args={'q': 'sha1:%s' % content['sha1']})
        expected_data = self.content_get_mimetype(content['sha1'])
        expected_data['content_url'] = content_url
        self.assertEqual(rv.data, expected_data)

    @given(unknown_content())
    def test_api_content_filetype_sha_not_found(self, unknown_content):

        url = reverse('api-content-filetype',
                      url_args={'q': 'sha1:%s' % unknown_content['sha1']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'No filetype information found for content '
            'sha1:%s.' % unknown_content['sha1']
        })

    @given(content())
    def test_api_content_language(self, content):

        self.content_add_language(content['sha1'])
        url = reverse('api-content-language',
                      url_args={'q': 'sha1_git:%s' % content['sha1_git']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        content_url = reverse('api-content',
                              url_args={'q': 'sha1:%s' % content['sha1']})
        expected_data = self.content_get_language(content['sha1'])
        expected_data['content_url'] = content_url
        self.assertEqual(rv.data, expected_data)

    @given(unknown_content())
    def test_api_content_language_sha_not_found(self, unknown_content):

        url = reverse('api-content-language',
                      url_args={'q': 'sha1:%s' % unknown_content['sha1']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'No language information found for content '
            'sha1:%s.' % unknown_content['sha1']
        })

    @pytest.mark.skipif(ctags_json_missing,
                        reason="requires ctags with json output support")
    @given(contents_with_ctags())
    def test_api_content_symbol(self, contents_with_ctags):

        expected_data = {}
        for content_sha1 in contents_with_ctags['sha1s']:
            self.content_add_ctags(content_sha1)
            for ctag in self.content_get_ctags(content_sha1):
                if ctag['name'] == contents_with_ctags['symbol_name']:
                    expected_data[content_sha1] = ctag
                    break
        url = reverse('api-content-symbol',
                      url_args={'q': contents_with_ctags['symbol_name']},
                      query_params={'per_page': 100})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        for entry in rv.data:
            content_sha1 = entry['sha1']
            expected_entry = expected_data[content_sha1]
            for key, view_name in (('content_url', 'api-content'),
                                   ('data_url', 'api-content-raw'),
                                   ('license_url', 'api-content-license'),
                                   ('language_url', 'api-content-language'),
                                   ('filetype_url', 'api-content-filetype')):
                expected_entry[key] = reverse(view_name,
                                              url_args={'q': 'sha1:%s' %
                                                        content_sha1})
            expected_entry['sha1'] = content_sha1
            del expected_entry['id']
            self.assertEqual(entry, expected_entry)
        self.assertFalse('Link' in rv)

        url = reverse('api-content-symbol',
                      url_args={'q': contents_with_ctags['symbol_name']},
                      query_params={'per_page': 2})
        rv = self.client.get(url)

        next_url = reverse('api-content-symbol',
                           url_args={'q': contents_with_ctags['symbol_name']},
                           query_params={'last_sha1': rv.data[1]['sha1'],
                                         'per_page': 2})
        self.assertEqual(rv['Link'], '<%s>; rel="next"' % next_url)

    @pytest.mark.xfail(reason='FIXME: exception should be raised in service')
    @given(unknown_content())
    def test_api_content_symbol_not_found(self, unknown_content):

        url = reverse('api-content-symbol', url_args={'q': 'bar'},
                      query_params={'last_sha1': 'hash'})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'No indexed raw content match expression \'bar\'.'
        })
        self.assertFalse('Link' in rv)

    @pytest.mark.skipif(ctags_json_missing,
                        reason="requires ctags with json output support")
    @given(content())
    def test_api_content_ctags(self, content):

        self.content_add_ctags(content['sha1'])
        url = reverse('api-content-ctags',
                      url_args={'q': 'sha1_git:%s' % content['sha1_git']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        content_url = reverse('api-content',
                              url_args={'q': 'sha1:%s' % content['sha1']})
        expected_data = list(self.content_get_ctags(content['sha1']))
        for e in expected_data:
            e['content_url'] = content_url
        self.assertEqual(rv.data, expected_data)

    @pytest.mark.skipif(fossology_missing,
                        reason="requires fossology-nomossa installed")
    @given(content())
    def test_api_content_license(self, content):

        self.content_add_license(content['sha1'])
        url = reverse('api-content-license',
                      url_args={'q': 'sha1_git:%s' % content['sha1_git']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        content_url = reverse('api-content',
                              url_args={'q': 'sha1:%s' % content['sha1']})
        expected_data = self.content_get_license(content['sha1'])
        expected_data['content_url'] = content_url
        self.assertEqual(rv.data, expected_data)

    @given(unknown_content())
    def test_api_content_license_sha_not_found(self, unknown_content):

        url = reverse('api-content-license',
                      url_args={'q': 'sha1:%s' % unknown_content['sha1']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'No license information found for content '
            'sha1:%s.' % unknown_content['sha1']
        })

    @given(content())
    def test_api_content_metadata(self, content):

        url = reverse('api-content', {'q': 'sha1:%s' % content['sha1']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        expected_data = self.content_get_metadata(content['sha1'])
        for key, view_name in (('data_url', 'api-content-raw'),
                               ('license_url', 'api-content-license'),
                               ('language_url', 'api-content-language'),
                               ('filetype_url', 'api-content-filetype')):
            expected_data[key] = reverse(view_name,
                                         url_args={'q': 'sha1:%s' %
                                                   content['sha1']})
        self.assertEqual(rv.data, expected_data)

    @given(unknown_content())
    def test_api_content_not_found_as_json(self, unknown_content):

        url = reverse('api-content',
                      url_args={'q': 'sha1:%s' % unknown_content['sha1']})
        rv = self.client.get(url)
        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'Content with sha1 checksum equals to %s not found!'
            % unknown_content['sha1']
        })

    @given(unknown_content())
    def test_api_content_not_found_as_yaml(self, unknown_content):

        url = reverse('api-content',
                      url_args={'q': 'sha256:%s' % unknown_content['sha256']})
        rv = self.client.get(url, HTTP_ACCEPT='application/yaml')

        self.assertEqual(rv.status_code, 404)
        self.assertTrue('application/yaml' in rv['Content-Type'])

        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'Content with sha256 checksum equals to %s not found!' %
            unknown_content['sha256']
        })

    @given(unknown_content())
    def test_api_content_raw_ko_not_found(self, unknown_content):

        url = reverse('api-content-raw',
                      url_args={'q': 'sha1:%s' % unknown_content['sha1']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'Content with sha1 checksum equals to %s not found!' %
            unknown_content['sha1']
        })

    @given(content())
    def test_api_content_raw_text(self, content):

        url = reverse('api-content-raw',
                      url_args={'q': 'sha1:%s' % content['sha1']})

        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/octet-stream')
        self.assertEqual(
            rv['Content-disposition'],
            'attachment; filename=content_sha1_%s_raw' % content['sha1'])
        self.assertEqual(
            rv['Content-Type'], 'application/octet-stream')
        expected_data = self.content_get(content['sha1'])
        self.assertEqual(rv.content, expected_data['data'])

    @given(content())
    def test_api_content_raw_text_with_filename(self, content):

        url = reverse('api-content-raw',
                      url_args={'q': 'sha1:%s' % content['sha1']},
                      query_params={'filename': 'filename.txt'})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/octet-stream')
        self.assertEqual(
            rv['Content-disposition'],
            'attachment; filename=filename.txt')
        self.assertEqual(
            rv['Content-Type'], 'application/octet-stream')
        expected_data = self.content_get(content['sha1'])
        self.assertEqual(rv.content, expected_data['data'])

    @given(content())
    def test_api_check_content_known(self, content):

        url = reverse('api-content-known',
                      url_args={'q': content['sha1']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')

        self.assertEqual(rv.data, {
            'search_res': [
                {
                    'found': True,
                    'sha1': content['sha1']
                }
            ],
            'search_stats': {'nbfiles': 1, 'pct': 100.0}
        })

    @given(content())
    def test_api_check_content_known_as_yaml(self, content):

        url = reverse('api-content-known',
                      url_args={'q': content['sha1']})
        rv = self.client.get(url, HTTP_ACCEPT='application/yaml')

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/yaml')

        self.assertEqual(rv.data, {
            'search_res': [
                {
                    'found': True,
                    'sha1': content['sha1']
                }
            ],
            'search_stats': {'nbfiles': 1, 'pct': 100.0}
        })

    @given(content())
    def test_api_check_content_known_post_as_yaml(self, content):

        url = reverse('api-content-known')
        rv = self.client.post(
            url,
            data={
                'q': content['sha1']
            },
            HTTP_ACCEPT='application/yaml'
        )

        self.assertEqual(rv.status_code, 200)
        self.assertTrue('application/yaml' in rv['Content-Type'])
        self.assertEqual(rv.data, {
            'search_res': [
                {
                    'found': True,
                    'sha1': content['sha1']
                }
            ],
            'search_stats': {'nbfiles': 1, 'pct': 100.0}
        })

    @given(unknown_content())
    def test_api_check_content_known_not_found(self, unknown_content):

        url = reverse('api-content-known',
                      url_args={'q': unknown_content['sha1']})
        rv = self.client.get(url)

        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'search_res': [
                {
                    'found': False,
                    'sha1': unknown_content['sha1']
                }
            ],
            'search_stats': {'nbfiles': 1, 'pct': 0.0}
        })



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.test import APITestCase
from unittest.mock import patch

from swh.web.tests.testcase import WebTestCase


class ReleaseApiTestCase(WebTestCase, APITestCase):

    @patch('swh.web.api.views.release.service')
    def test_api_release(self, mock_service):
        release_id = '7045404f3d1c54e6473'
        target_id = '6072557b6c10cd9a211'
        # given
        stub_release = {
            'id': release_id,
            'target_type': 'revision',
            'target': target_id,
            "date": "Mon, 10 Mar 1997 08:00:00 GMT",
            "synthetic": True,
            'author': {
                'id': 10,
                'name': 'author release name',
                'email': 'author@email',
            },
        }

        expected_release = {
            'id': release_id,
            'target_type': 'revision',
            'target': target_id,
            'target_url': '/api/1/revision/%s/' % target_id,
            "date": "Mon, 10 Mar 1997 08:00:00 GMT",
            "synthetic": True,
            'author_url': '/api/1/person/10/',
            'author': {
                'id': 10,
                'name': 'author release name',
                'email': 'author@email',
            },
        }

        mock_service.lookup_release.return_value = stub_release

        # when
        rv = self.client.get('/api/1/release/%s/' % release_id)

        # then
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, expected_release)

        mock_service.lookup_release.assert_called_once_with(release_id)

    @patch('swh.web.api.views.release.service')
    def test_api_release_target_type_not_a_revision(self, mock_service):
        release = '8d56a78'
        target = '9a5c3f'
        # given
        stub_release = {
            'id': release,
            'target_type': 'other-stuff',
            'target': target,
            "date": "Mon, 10 Mar 1997 08:00:00 GMT",
            "synthetic": True,
            'author': {
                'id': 9,
                'name': 'author release name',
                'email': 'author@email',
            },
        }

        expected_release = {
            'id': release,
            'target_type': 'other-stuff',
            'target': target,
            "date": "Mon, 10 Mar 1997 08:00:00 GMT",
            "synthetic": True,
            'author_url': '/api/1/person/9/',
            'author': {
                'id': 9,
                'name': 'author release name',
                'email': 'author@email',
            },
        }

        mock_service.lookup_release.return_value = stub_release

        # when
        rv = self.client.get('/api/1/release/%s/' % release)

        # then
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, expected_release)

        mock_service.lookup_release.assert_called_once_with(release)

    @patch('swh.web.api.views.release.service')
    def test_api_release_not_found(self, mock_service):
        # given
        mock_service.lookup_release.return_value = None

        # when
        rv = self.client.get('/api/1/release/c54e6473c71bbb716529/')

        # then
        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'Release with sha1_git c54e6473c71bbb716529 not found.'
        })



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from distutils.util import strtobool

from swh.web.common import service
from swh.web.common.exc import BadInputExc
from swh.web.common.origin_visits import get_origin_visits
from swh.web.common.utils import reverse
from swh.web.api.apidoc import api_doc
from swh.web.api.apiurls import api_route
from swh.web.api.views.utils import api_lookup


def _enrich_origin(origin):
    if 'id' in origin:
        o = origin.copy()
        o['origin_visits_url'] = \
            reverse('api-origin-visits', url_args={'origin_id': origin['id']})
        return o

    return origin


@api_route(r'/origin/(?P<origin_id>[0-9]+)/', 'api-origin')
@api_route(r'/origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/',
           'api-origin')
@api_doc('/origin/')
def api_origin(request, origin_id=None, origin_type=None, origin_url=None):
    """
    .. http:get:: /api/1/origin/(origin_id)/

        Get information about a software origin.

        :param int origin_id: a software origin identifier

        :>json number id: the origin unique identifier
        :>json string origin_visits_url: link to in order to get information about the
            visits for that origin
        :>json string type: the type of software origin (possible values are ``git``, ``svn``,
            ``hg``, ``deb``, ``pypi``, ``ftp`` or ``deposit``)
        :>json string url: the origin canonical url

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested origin can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/1/`

    .. http:get:: /api/1/origin/(origin_type)/url/(origin_url)/

        Get information about a software origin.

        :param string origin_type: the origin type (possible values are ``git``, ``svn``,
            ``hg``, ``deb``, ``pypi``, ``ftp`` or ``deposit``)
        :param string origin_url: the origin url

        :>json number id: the origin unique identifier
        :>json string origin_visits_url: link to in order to get information about the
            visits for that origin
        :>json string type: the type of software origin
        :>json string url: the origin canonical url

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested origin can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/git/url/https://github.com/python/cpython/`
    """ # noqa
    ori_dict = {
        'id': origin_id,
        'type': origin_type,
        'url': origin_url
    }
    ori_dict = {k: v for k, v in ori_dict.items() if ori_dict[k]}
    if 'id' in ori_dict:
        error_msg = 'Origin with id %s not found.' % ori_dict['id']
    else:
        error_msg = 'Origin with type %s and URL %s not found' % (
            ori_dict['type'], ori_dict['url'])

    return api_lookup(
        service.lookup_origin, ori_dict,
        notfound_msg=error_msg,
        enrich_fn=_enrich_origin)


@api_route(r'/origin/search/(?P<url_pattern>.+)/',
           'api-origin-search')
@api_doc('/origin/search/')
def api_origin_search(request, url_pattern):
    """
    .. http:get:: /api/1/origin/search/(url_pattern)/

        Search for software origins whose urls contain a provided string
        pattern or match a provided regular expression.
        The search is performed in a case insensitive way.

        :param string url_pattern: a string pattern or a regular expression
        :query int offset: the number of found origins to skip before returning results
        :query int limit: the maximum number of found origins to return
        :query boolean regexp: if true, consider provided pattern as a regular expression
            and search origins whose urls match it
        :query boolean with_visit: if true, only return origins with at least one visit
            by Software heritage

        :>jsonarr number id: the origin unique identifier
        :>jsonarr string origin_visits_url: link to in order to get information about the
            visits for that origin
        :>jsonarr string type: the type of software origin
        :>jsonarr string url: the origin canonical url

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/search/python/?limit=2`
    """ # noqa
    result = {}
    offset = int(request.query_params.get('offset', '0'))
    limit = int(request.query_params.get('limit', '70'))
    regexp = request.query_params.get('regexp', 'false')
    with_visit = request.query_params.get('with_visit', 'false')

    results = api_lookup(service.search_origin, url_pattern, offset, limit,
                         bool(strtobool(regexp)), bool(strtobool(with_visit)),
                         enrich_fn=_enrich_origin)

    nb_results = len(results)
    if nb_results == limit:
        query_params = {}
        query_params['offset'] = offset + limit
        query_params['limit'] = limit
        query_params['regexp'] = regexp

        result['headers'] = {
            'link-next': reverse('api-origin-search',
                                 url_args={'url_pattern': url_pattern},
                                 query_params=query_params)
        }

    result.update({
        'results': results
    })

    return result


@api_route(r'/origin/metadata-search/',
           'api-origin-metadata-search')
@api_doc('/origin/metadata-search/', noargs=True)
def api_origin_metadata_search(request):
    """
    .. http:get:: /api/1/origin/metadata-search/

        Search for software origins whose metadata (expressed as a
        JSON-LD/CodeMeta dictionary) match the provided criteria.
        For now, only full-text search on this dictionary is supported.

        :query str fulltext: a string that will be matched against origin metadata;
            results are ranked and ordered starting with the best ones.
        :query int limit: the maximum number of found origins to return
            (bounded to 100)

        :>jsonarr number origin_id: the origin unique identifier
        :>jsonarr dict metadata: metadata of the origin (as a JSON-LD/CodeMeta dictionary)
        :>jsonarr string from_revision: the revision used to extract these
            metadata (the current HEAD or one of the former HEADs)
        :>jsonarr dict tool: the tool used to extract these metadata

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/metadata-search/?limit=2&fulltext=Jane%20Doe`
    """ # noqa
    fulltext = request.query_params.get('fulltext', None)
    limit = min(int(request.query_params.get('limit', '70')), 100)

    if not fulltext:
        content = '"fulltext" must be provided and non-empty.'
        raise BadInputExc(content)

    results = api_lookup(service.search_origin_metadata, fulltext, limit)

    return {
        'results': results,
    }


@api_route(r'/origin/(?P<origin_id>[0-9]+)/visits/', 'api-origin-visits')
@api_doc('/origin/visits/')
def api_origin_visits(request, origin_id):
    """
    .. http:get:: /api/1/origin/(origin_id)/visits/

        Get information about all visits of a software origin.
        Visits are returned sorted in descending order according
        to their date.

        :param int origin_id: a software origin identifier
        :query int per_page: specify the number of visits to list, for pagination purposes
        :query int last_visit: visit to start listing from, for pagination purposes

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request
        :resheader Link: indicates that a subsequent result page is available and contains
            the url pointing to it

        :>jsonarr string date: ISO representation of the visit date (in UTC)
        :>jsonarr number id: the unique identifier of the origin
        :>jsonarr string origin_visit_url: link to :http:get:`/api/1/origin/(origin_id)/visit/(visit_id)/`
            in order to get information about the visit
        :>jsonarr string snapshot: the snapshot identifier of the visit
        :>jsonarr string snapshot_url: link to :http:get:`/api/1/snapshot/(snapshot_id)/`
            in order to get information about the snapshot of the visit
        :>jsonarr string status: status of the visit (either **full**, **partial** or **ongoing**)
        :>jsonarr number visit: the unique identifier of the visit

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested origin can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/1/visits/`
    """ # noqa
    result = {}
    per_page = int(request.query_params.get('per_page', '10'))
    last_visit = request.query_params.get('last_visit')
    if last_visit:
        last_visit = int(last_visit)

    def _lookup_origin_visits(
            origin_id, last_visit=last_visit, per_page=per_page):
        all_visits = get_origin_visits({'id': origin_id})
        all_visits.reverse()
        visits = []
        if not last_visit:
            visits = all_visits[:per_page]
        else:
            for i, v in enumerate(all_visits):
                if v['visit'] == last_visit:
                    visits = all_visits[i+1:i+1+per_page]
                    break
        for v in visits:
            yield v

    def _enrich_origin_visit(origin_visit):
        ov = origin_visit.copy()
        ov['origin_visit_url'] = reverse('api-origin-visit',
                                         url_args={'origin_id': origin_id,
                                                   'visit_id': ov['visit']})
        snapshot = ov['snapshot']
        if snapshot:
            ov['snapshot_url'] = reverse('api-snapshot',
                                         url_args={'snapshot_id': snapshot})
        else:
            ov['snapshot_url'] = None
        return ov

    results = api_lookup(_lookup_origin_visits, origin_id,
                         notfound_msg='No origin {} found'.format(origin_id),
                         enrich_fn=_enrich_origin_visit)

    if results:
        nb_results = len(results)
        if nb_results == per_page:
            new_last_visit = results[-1]['visit']
            query_params = {}
            query_params['last_visit'] = new_last_visit

            if request.query_params.get('per_page'):
                query_params['per_page'] = per_page

            result['headers'] = {
                'link-next': reverse('api-origin-visits',
                                     url_args={'origin_id': origin_id},
                                     query_params=query_params)
            }

    result.update({
        'results': results
    })

    return result


@api_route(r'/origin/(?P<origin_id>[0-9]+)/visit/(?P<visit_id>[0-9]+)/',
           'api-origin-visit')
@api_doc('/origin/visit/')
def api_origin_visit(request, origin_id, visit_id):
    """
    .. http:get:: /api/1/origin/(origin_id)/visit/(visit_id)/

        Get information about a specific visit of a software origin.

        :param int origin_id: a software origin identifier
        :param int visit_id: a visit identifier

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        :>json string date: ISO representation of the visit date (in UTC)
        :>json number origin: the origin unique identifier
        :>json string origin_url: link to get information about the origin
        :>jsonarr string snapshot: the snapshot identifier of the visit
        :>jsonarr string snapshot_url: link to :http:get:`/api/1/snapshot/(snapshot_id)/`
            in order to get information about the snapshot of the visit
        :>json string status: status of the visit (either **full**, **partial** or **ongoing**)
        :>json number visit: the unique identifier of the visit

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested origin or visit can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`origin/1500/visit/1/`
    """ # noqa
    def _enrich_origin_visit(origin_visit):
        ov = origin_visit.copy()
        ov['origin_url'] = reverse('api-origin',
                                   url_args={'origin_id': ov['origin']})
        snapshot = ov['snapshot']
        if snapshot:
            ov['snapshot_url'] = reverse('api-snapshot',
                                         url_args={'snapshot_id': snapshot})
        else:
            ov['snapshot_url'] = None

        return ov

    return api_lookup(
        service.lookup_origin_visit, origin_id, visit_id,
        notfound_msg=('No visit {} for origin {} found'
                      .format(visit_id, origin_id)),
        enrich_fn=_enrich_origin_visit)



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch, call

from swh.web.api import utils
from swh.web.tests.testcase import WebTestCase


class UtilsTestCase(WebTestCase):
    def setUp(self):
        self.maxDiff = None
        self.url_map = [dict(rule='/other/<slug>',
                             methods=set(['GET', 'POST', 'HEAD']),
                             endpoint='foo'),
                        dict(rule='/some/old/url/<slug>',
                             methods=set(['GET', 'POST']),
                             endpoint='blablafn'),
                        dict(rule='/other/old/url/<int:id>',
                             methods=set(['GET', 'HEAD']),
                             endpoint='bar'),
                        dict(rule='/other',
                             methods=set([]),
                             endpoint=None),
                        dict(rule='/other2',
                             methods=set([]),
                             endpoint=None)]
        self.sample_content_hashes = {
            'blake2s256': ('791e07fcea240ade6dccd0a9309141673'
                           'c31242cae9c237cf3855e151abc78e9'),
            'sha1': 'dc2830a9e72f23c1dfebef4413003221baa5fb62',
            'sha1_git': 'fe95a46679d128ff167b7c55df5d02356c5a1ae1',
            'sha256': ('b5c7fe0536f44ef60c8780b6065d30bca74a5cd06'
                       'd78a4a71ba1ad064770f0c9')
        }

    def test_filter_field_keys_dict_unknown_keys(self):
        # when
        actual_res = utils.filter_field_keys(
            {'directory': 1, 'file': 2, 'link': 3},
            {'directory1', 'file2'})

        # then
        self.assertEqual(actual_res, {})

    def test_filter_field_keys_dict(self):
        # when
        actual_res = utils.filter_field_keys(
            {'directory': 1, 'file': 2, 'link': 3},
            {'directory', 'link'})

        # then
        self.assertEqual(actual_res, {'directory': 1, 'link': 3})

    def test_filter_field_keys_list_unknown_keys(self):
        # when
        actual_res = utils.filter_field_keys(
            [{'directory': 1, 'file': 2, 'link': 3},
             {'1': 1, '2': 2, 'link': 3}],
            {'d'})

        # then
        self.assertEqual(actual_res, [{}, {}])

    def test_filter_field_keys_map(self):
        # when
        actual_res = utils.filter_field_keys(
            map(lambda x: {'i': x['i']+1, 'j': x['j']},
                [{'i': 1, 'j': None},
                 {'i': 2, 'j': None},
                 {'i': 3, 'j': None}]),
            {'i'})

        # then
        self.assertEqual(list(actual_res), [{'i': 2}, {'i': 3}, {'i': 4}])

    def test_filter_field_keys_list(self):
        # when
        actual_res = utils.filter_field_keys(
            [{'directory': 1, 'file': 2, 'link': 3},
             {'dir': 1, 'fil': 2, 'lin': 3}],
            {'directory', 'dir'})

        # then
        self.assertEqual(actual_res, [{'directory': 1}, {'dir': 1}])

    def test_filter_field_keys_other(self):
        # given
        input_set = {1, 2}

        # when
        actual_res = utils.filter_field_keys(input_set, {'a', '1'})

        # then
        self.assertEqual(actual_res, input_set)

    def test_person_to_string(self):
        self.assertEqual(utils.person_to_string(dict(name='raboof',
                                                     email='foo@bar')),
                         'raboof <foo@bar>')

    def test_enrich_release_0(self):
        # when
        actual_release = utils.enrich_release({})

        # then
        self.assertEqual(actual_release, {})

    @patch('swh.web.api.utils.reverse')
    def test_enrich_release_1(self, mock_django_reverse):
        # given

        def reverse_test_context(view_name, url_args):
            if view_name == 'api-content':
                id = url_args['q']
                return '/api/1/content/%s/' % id
            elif view_name == 'api-person':
                id = url_args['person_id']
                return '/api/1/person/%s/' % id
            else:
                raise ValueError(
                    'This should not happened so fail if it does.')

        mock_django_reverse.side_effect = reverse_test_context

        # when
        actual_release = utils.enrich_release({
            'target': '123',
            'target_type': 'content',
            'author': {
                'id': 100,
                'name': 'author release name',
                'email': 'author@email',
            },
        })

        # then
        self.assertEqual(actual_release, {
            'target': '123',
            'target_type': 'content',
            'target_url': '/api/1/content/sha1_git:123/',
            'author_url': '/api/1/person/100/',
            'author': {
                'id': 100,
                'name': 'author release name',
                'email': 'author@email',
            },
        })

        mock_django_reverse.assert_has_calls([
                call('api-content', url_args={'q': 'sha1_git:123'}),
                call('api-person', url_args={'person_id': 100})
        ])

    @patch('swh.web.api.utils.reverse')
    def test_enrich_release_2(self, mock_django_reverse):
        # given
        mock_django_reverse.return_value = '/api/1/dir/23/'

        # when
        actual_release = utils.enrich_release({'target': '23',
                                               'target_type': 'directory'})

        # then
        self.assertEqual(actual_release, {
            'target': '23',
            'target_type': 'directory',
            'target_url': '/api/1/dir/23/'
        })

        mock_django_reverse.assert_called_once_with('api-directory',
                                                    url_args={'sha1_git': '23'}) # noqa

    @patch('swh.web.api.utils.reverse')
    def test_enrich_release_3(self, mock_django_reverse):
        # given
        mock_django_reverse.return_value = '/api/1/rev/3/'

        # when
        actual_release = utils.enrich_release({'target': '3',
                                               'target_type': 'revision'})

        # then
        self.assertEqual(actual_release, {
            'target': '3',
            'target_type': 'revision',
            'target_url': '/api/1/rev/3/'
        })

        mock_django_reverse.assert_called_once_with('api-revision',
                                                    url_args={'sha1_git': '3'})

    @patch('swh.web.api.utils.reverse')
    def test_enrich_release_4(self, mock_django_reverse):
        # given
        mock_django_reverse.return_value = '/api/1/rev/4/'

        # when
        actual_release = utils.enrich_release({'target': '4',
                                               'target_type': 'release'})

        # then
        self.assertEqual(actual_release, {
            'target': '4',
            'target_type': 'release',
            'target_url': '/api/1/rev/4/'
        })

        mock_django_reverse.assert_called_once_with('api-release',
                                                    url_args={'sha1_git': '4'})

    @patch('swh.web.api.utils.reverse')
    def test_enrich_directory_no_type(self, mock_django_reverse):
        # when/then
        self.assertEqual(utils.enrich_directory({'id': 'dir-id'}),
                         {'id': 'dir-id'})

        # given
        mock_django_reverse.return_value = '/api/content/sha1_git:123/'

        # when
        actual_directory = utils.enrich_directory({
            'id': 'dir-id',
            'type': 'file',
            'target': '123',
        })

        # then
        self.assertEqual(actual_directory, {
            'id': 'dir-id',
            'type': 'file',
            'target': '123',
            'target_url': '/api/content/sha1_git:123/',
        })

        mock_django_reverse.assert_called_once_with(
            'api-content', url_args={'q': 'sha1_git:123'})

    @patch('swh.web.api.utils.reverse')
    def test_enrich_directory_with_context_and_type_file(
        self, mock_django_reverse,
    ):
        # given
        mock_django_reverse.return_value = '/api/content/sha1_git:123/'

        # when
        actual_directory = utils.enrich_directory({
            'id': 'dir-id',
            'type': 'file',
            'name': 'hy',
            'target': '789',
        }, context_url='/api/revision/revsha1/directory/prefix/path/')

        # then
        self.assertEqual(actual_directory, {
            'id': 'dir-id',
            'type': 'file',
            'name': 'hy',
            'target': '789',
            'target_url': '/api/content/sha1_git:123/',
            'file_url': '/api/revision/revsha1/directory'
                        '/prefix/path/hy/'
        })

        mock_django_reverse.assert_called_once_with(
            'api-content', url_args={'q': 'sha1_git:789'})

    @patch('swh.web.api.utils.reverse')
    def test_enrich_directory_with_context_and_type_dir(
        self, mock_django_reverse,
    ):
        # given
        mock_django_reverse.return_value = '/api/directory/456/'

        # when
        actual_directory = utils.enrich_directory({
            'id': 'dir-id',
            'type': 'dir',
            'name': 'emacs-42',
            'target_type': 'file',
            'target': '456',
        }, context_url='/api/revision/origin/2/directory/some/prefix/path/')

        # then
        self.assertEqual(actual_directory, {
            'id': 'dir-id',
            'type': 'dir',
            'target_type': 'file',
            'name': 'emacs-42',
            'target': '456',
            'target_url': '/api/directory/456/',
            'dir_url': '/api/revision/origin/2/directory'
            '/some/prefix/path/emacs-42/'
        })

        mock_django_reverse.assert_called_once_with('api-directory',
                                                    url_args={'sha1_git': '456'}) # noqa

    def test_enrich_content_without_hashes(self):
        # when/then
        self.assertEqual(utils.enrich_content({'id': '123'}),
                         {'id': '123'})

    @patch('swh.web.api.utils.reverse')
    def test_enrich_content_with_hashes(self, mock_django_reverse):

        for algo, hash in self.sample_content_hashes.items():

            query_string = '%s:%s' % (algo, hash)

            # given
            mock_django_reverse.side_effect = [
                '/api/content/%s/raw/' % query_string,
                '/api/filetype/%s/' % query_string,
                '/api/language/%s/' % query_string,
                '/api/license/%s/' % query_string
                ]

            # when
            enriched_content = utils.enrich_content(
                {
                    algo: hash,
                },
                query_string=query_string
            )

            # then
            self.assertEqual(
                enriched_content,
                {
                    algo: hash,
                    'data_url': '/api/content/%s/raw/' % query_string,
                    'filetype_url': '/api/filetype/%s/' % query_string,
                    'language_url': '/api/language/%s/' % query_string,
                    'license_url': '/api/license/%s/' % query_string,
                }
            )

            mock_django_reverse.assert_has_calls([
                call('api-content-raw', url_args={'q': query_string}),
                call('api-content-filetype', url_args={'q': query_string}),
                call('api-content-language', url_args={'q': query_string}),
                call('api-content-license', url_args={'q': query_string}),
            ])

            mock_django_reverse.reset()

    @patch('swh.web.api.utils.reverse')
    def test_enrich_content_with_hashes_and_top_level_url(self,
                                                          mock_django_reverse):

        for algo, hash in self.sample_content_hashes.items():

            query_string = '%s:%s' % (algo, hash)

            # given
            mock_django_reverse.side_effect = [
                '/api/content/%s/' % query_string,
                '/api/content/%s/raw/' % query_string,
                '/api/filetype/%s/' % query_string,
                '/api/language/%s/' % query_string,
                '/api/license/%s/' % query_string,
                ]

            # when
            enriched_content = utils.enrich_content(
                {
                    algo: hash
                },
                top_url=True,
                query_string=query_string
            )

            # then
            self.assertEqual(
                enriched_content,
                {
                    algo: hash,
                    'content_url': '/api/content/%s/' % query_string,
                    'data_url': '/api/content/%s/raw/' % query_string,
                    'filetype_url': '/api/filetype/%s/' % query_string,
                    'language_url': '/api/language/%s/' % query_string,
                    'license_url': '/api/license/%s/' % query_string,
                }
            )

            mock_django_reverse.assert_has_calls([
                call('api-content', url_args={'q': query_string}),
                call('api-content-raw', url_args={'q': query_string}),
                call('api-content-filetype', url_args={'q': query_string}),
                call('api-content-language', url_args={'q': query_string}),
                call('api-content-license', url_args={'q': query_string}),
            ])

            mock_django_reverse.reset()

    def _reverse_context_test(self, view_name, url_args):
        if view_name == 'api-revision':
            return '/api/revision/%s/' % url_args['sha1_git']
        elif view_name == 'api-revision-context':
            return '/api/revision/%s/prev/%s/' % (url_args['sha1_git'], url_args['context'])  # noqa
        elif view_name == 'api-revision-log':
            if 'prev_sha1s' in url_args:
                return '/api/revision/%s/prev/%s/log/' % (url_args['sha1_git'], url_args['prev_sha1s'])  # noqa
            else:
                return '/api/revision/%s/log/' % url_args['sha1_git']

    @patch('swh.web.api.utils.reverse')
    def test_enrich_revision_without_children_or_parent(
        self, mock_django_reverse,
    ):
        # given
        def reverse_test(view_name, url_args):
            if view_name == 'api-revision':
                return '/api/revision/' + url_args['sha1_git'] + '/'
            elif view_name == 'api-revision-log':
                return '/api/revision/' + url_args['sha1_git'] + '/log/'
            elif view_name == 'api-directory':
                return '/api/directory/' + url_args['sha1_git'] + '/'
            elif view_name == 'api-person':
                return '/api/person/' + url_args['person_id'] + '/'

        mock_django_reverse.side_effect = reverse_test

        # when
        actual_revision = utils.enrich_revision({
            'id': 'rev-id',
            'directory': '123',
            'author': {'id': '1'},
            'committer': {'id': '2'},
        })

        expected_revision = {
            'id': 'rev-id',
            'directory': '123',
            'url': '/api/revision/rev-id/',
            'history_url': '/api/revision/rev-id/log/',
            'directory_url': '/api/directory/123/',
            'author': {'id': '1'},
            'author_url': '/api/person/1/',
            'committer': {'id': '2'},
            'committer_url': '/api/person/2/'
        }

        # then
        self.assertEqual(actual_revision, expected_revision)

        mock_django_reverse.assert_has_calls(
            [call('api-revision', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id'}),
             call('api-person', url_args={'person_id': '1'}),
             call('api-person', url_args={'person_id': '2'}),
             call('api-directory', url_args={'sha1_git': '123'})])

    @patch('swh.web.api.utils.reverse')
    def test_enrich_revision_with_children_and_parent_no_dir(
        self, mock_django_reverse,
    ):
        # given
        mock_django_reverse.side_effect = self._reverse_context_test

        # when
        actual_revision = utils.enrich_revision({
            'id': 'rev-id',
            'parents': ['123'],
            'children': ['456'],
        }, context='prev-rev')

        expected_revision = {
            'id': 'rev-id',
            'url': '/api/revision/rev-id/',
            'history_url': '/api/revision/rev-id/log/',
            'history_context_url': '/api/revision/rev-id/prev/prev-rev/log/',
            'parents': [{'id': '123', 'url': '/api/revision/123/'}],
            'children': ['456'],
            'children_urls': ['/api/revision/456/',
                              '/api/revision/prev-rev/'],
        }

        # then
        self.assertEqual(actual_revision, expected_revision)

        mock_django_reverse.assert_has_calls(
            [call('api-revision', url_args={'sha1_git': 'prev-rev'}),
             call('api-revision', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id',
                                                'prev_sha1s': 'prev-rev'}),
             call('api-revision', url_args={'sha1_git': '123'}),
             call('api-revision', url_args={'sha1_git': '456'})])

    @patch('swh.web.api.utils.reverse')
    def test_enrich_revision_no_context(self, mock_django_reverse):
        # given
        mock_django_reverse.side_effect = self._reverse_context_test

        # when
        actual_revision = utils.enrich_revision({
            'id': 'rev-id',
            'parents': ['123'],
            'children': ['456'],
        })

        expected_revision = {
            'id': 'rev-id',
            'url': '/api/revision/rev-id/',
            'history_url': '/api/revision/rev-id/log/',
            'parents': [{'id': '123', 'url': '/api/revision/123/'}],
            'children': ['456'],
            'children_urls': ['/api/revision/456/']
        }

        # then
        self.assertEqual(actual_revision, expected_revision)

        mock_django_reverse.assert_has_calls(
            [call('api-revision', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id'}),
             call('api-revision', url_args={'sha1_git': '123'}),
             call('api-revision', url_args={'sha1_git': '456'})])

    @patch('swh.web.api.utils.reverse')
    def test_enrich_revision_context_empty_prev_list(
        self, mock_django_reverse,
    ):
        # given
        mock_django_reverse.side_effect = self._reverse_context_test

        # when
        expected_revision = {
            'id': 'rev-id',
            'url': '/api/revision/rev-id/',
            'history_url': '/api/revision/rev-id/log/',
            'history_context_url': ('/api/revision/rev-id/'
                                    'prev/prev-rev/log/'),
            'parents': [{'id': '123', 'url': '/api/revision/123/'}],
            'children': ['456'],
            'children_urls': ['/api/revision/456/',
                              '/api/revision/prev-rev/'],
        }

        actual_revision = utils.enrich_revision({
            'id': 'rev-id',
            'url': '/api/revision/rev-id/',
            'parents': ['123'],
            'children': ['456']}, context='prev-rev')

        # then
        self.assertEqual(actual_revision, expected_revision)
        mock_django_reverse.assert_has_calls(
            [call('api-revision', url_args={'sha1_git': 'prev-rev'}),
             call('api-revision', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id',
                                                'prev_sha1s': 'prev-rev'}),
             call('api-revision', url_args={'sha1_git': '123'}),
             call('api-revision', url_args={'sha1_git': '456'})])

    @patch('swh.web.api.utils.reverse')
    def test_enrich_revision_context_some_prev_list(self, mock_django_reverse):
        # given
        mock_django_reverse.side_effect = self._reverse_context_test

        # when
        expected_revision = {
            'id': 'rev-id',
            'url': '/api/revision/rev-id/',
            'history_url': '/api/revision/rev-id/log/',
            'history_context_url': ('/api/revision/rev-id/'
                                    'prev/prev1-rev/prev0-rev/log/'),
            'parents': [{'id': '123', 'url': '/api/revision/123/'}],
            'children': ['456'],
            'children_urls': ['/api/revision/456/',
                              '/api/revision/prev0-rev/prev/prev1-rev/'],
        }

        actual_revision = utils.enrich_revision({
            'id': 'rev-id',
            'parents': ['123'],
            'children': ['456']}, context='prev1-rev/prev0-rev')

        # then
        self.assertEqual(actual_revision, expected_revision)
        mock_django_reverse.assert_has_calls(
            [call('api-revision-context', url_args={'context': 'prev1-rev',
                                                  'sha1_git': 'prev0-rev'}),
             call('api-revision', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'prev_sha1s': 'prev1-rev/prev0-rev', # noqa
                                              'sha1_git': 'rev-id'}),
             call('api-revision', url_args={'sha1_git': '123'}),
             call('api-revision', url_args={'sha1_git': '456'})])

    def _reverse_rev_message_test(self, view_name, url_args):
        if view_name == 'api-revision':
            return '/api/revision/%s/' % url_args['sha1_git']
        elif view_name == 'api-revision-log':
            if 'prev_sha1s' in url_args and url_args['prev_sha1s'] is not None:
                return '/api/revision/%s/prev/%s/log/' % (url_args['sha1_git'], url_args['prev_sha1s'])  # noqa
            else:
                return '/api/revision/%s/log/' % url_args['sha1_git']
        elif view_name == 'api-revision-raw-message':
            return '/api/revision/' + url_args['sha1_git'] + '/raw/'
        else:
            return '/api/revision/%s/prev/%s/' % (url_args['sha1_git'], url_args['context'])  # noqa

    @patch('swh.web.api.utils.reverse')
    def test_enrich_revision_with_no_message(self, mock_django_reverse):
        # given
        mock_django_reverse.side_effect = self._reverse_rev_message_test

        # when
        expected_revision = {
            'id': 'rev-id',
            'url': '/api/revision/rev-id/',
            'history_url': '/api/revision/rev-id/log/',
            'history_context_url': ('/api/revision/rev-id/'
                                    'prev/prev-rev/log/'),
            'message': None,
            'parents': [{'id': '123', 'url': '/api/revision/123/'}],
            'children': ['456'],
            'children_urls': ['/api/revision/456/',
                              '/api/revision/prev-rev/'],
        }

        actual_revision = utils.enrich_revision({
            'id': 'rev-id',
            'message': None,
            'parents': ['123'],
            'children': ['456'],
        }, context='prev-rev')

        # then
        self.assertEqual(actual_revision, expected_revision)

        mock_django_reverse.assert_has_calls(
            [call('api-revision', url_args={'sha1_git': 'prev-rev'}),
             call('api-revision', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id',
                                                'prev_sha1s': 'prev-rev'}),
             call('api-revision', url_args={'sha1_git': '123'}),
             call('api-revision', url_args={'sha1_git': '456'})]
        )

    @patch('swh.web.api.utils.reverse')
    def test_enrich_revision_with_invalid_message(self, mock_django_reverse):
        # given
        mock_django_reverse.side_effect = self._reverse_rev_message_test

        # when
        actual_revision = utils.enrich_revision({
            'id': 'rev-id',
            'message': None,
            'message_decoding_failed': True,
            'parents': ['123'],
            'children': ['456'],
        }, context='prev-rev')

        expected_revision = {
            'id': 'rev-id',
            'url': '/api/revision/rev-id/',
            'history_url': '/api/revision/rev-id/log/',
            'history_context_url': ('/api/revision/rev-id/'
                                    'prev/prev-rev/log/'),
            'message': None,
            'message_decoding_failed': True,
            'message_url': '/api/revision/rev-id/raw/',
            'parents': [{'id': '123', 'url': '/api/revision/123/'}],
            'children': ['456'],
            'children_urls': ['/api/revision/456/',
                              '/api/revision/prev-rev/'],
        }

        # then
        self.assertEqual(actual_revision, expected_revision)

        mock_django_reverse.assert_has_calls(
            [call('api-revision', url_args={'sha1_git': 'prev-rev'}),
             call('api-revision', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id'}),
             call('api-revision-log', url_args={'sha1_git': 'rev-id',
                                                'prev_sha1s': 'prev-rev'}),
             call('api-revision', url_args={'sha1_git': '123'}),
             call('api-revision', url_args={'sha1_git': '456'}),
             call('api-revision-raw-message', url_args={'sha1_git': 'rev-id'})]) # noqa



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch

from swh.model import hashutil

from swh.web.common import query
from swh.web.common.exc import BadInputExc
from swh.web.tests.testcase import WebTestCase


class QueryTestCase(WebTestCase):
    def test_parse_hash_malformed_query_with_more_than_2_parts(self):
        with self.assertRaises(BadInputExc):
            query.parse_hash('sha1:1234567890987654:other-stuff')

    def test_parse_hash_guess_sha1(self):
        h = 'f1d2d2f924e986ac86fdf7b36c94bcdf32beec15'
        r = query.parse_hash(h)
        self.assertEqual(r, ('sha1', hashutil.hash_to_bytes(h)))

    def test_parse_hash_guess_sha256(self):
        h = '084C799CD551DD1D8D5C5F9A5D593B2' \
            'E931F5E36122ee5c793c1d08a19839cc0'
        r = query.parse_hash(h)
        self.assertEqual(r, ('sha256', hashutil.hash_to_bytes(h)))

    def test_parse_hash_guess_algo_malformed_hash(self):
        with self.assertRaises(BadInputExc):
            query.parse_hash('1234567890987654')

    def test_parse_hash_check_sha1(self):
        h = 'f1d2d2f924e986ac86fdf7b36c94bcdf32beec15'
        r = query.parse_hash('sha1:' + h)
        self.assertEqual(r, ('sha1', hashutil.hash_to_bytes(h)))

    def test_parse_hash_check_sha1_git(self):
        h = 'e1d2d2f924e986ac86fdf7b36c94bcdf32beec15'
        r = query.parse_hash('sha1_git:' + h)
        self.assertEqual(r, ('sha1_git', hashutil.hash_to_bytes(h)))

    def test_parse_hash_check_sha256(self):
        h = '084C799CD551DD1D8D5C5F9A5D593B2E931F5E36122ee5c793c1d08a19839cc0'
        r = query.parse_hash('sha256:' + h)
        self.assertEqual(r, ('sha256', hashutil.hash_to_bytes(h)))

    def test_parse_hash_check_algo_malformed_sha1_hash(self):
        with self.assertRaises(BadInputExc):
            query.parse_hash('sha1:1234567890987654')

    def test_parse_hash_check_algo_malformed_sha1_git_hash(self):
        with self.assertRaises(BadInputExc):
            query.parse_hash('sha1_git:1234567890987654')

    def test_parse_hash_check_algo_malformed_sha256_hash(self):
        with self.assertRaises(BadInputExc):
            query.parse_hash('sha256:1234567890987654')

    def test_parse_hash_check_algo_unknown_one(self):
        with self.assertRaises(BadInputExc):
            query.parse_hash('sha2:1234567890987654')

    @patch('swh.web.common.query.parse_hash')
    def test_parse_hash_with_algorithms_or_throws_bad_query(self, mock_hash):
        # given
        mock_hash.side_effect = BadInputExc('Error input')

        # when
        with self.assertRaises(BadInputExc) as cm:
            query.parse_hash_with_algorithms_or_throws(
                'sha1:blah',
                ['sha1'],
                'useless error message for this use case')
        self.assertIn('Error input', cm.exception.args[0])

        mock_hash.assert_called_once_with('sha1:blah')

    @patch('swh.web.common.query.parse_hash')
    def test_parse_hash_with_algorithms_or_throws_bad_algo(self, mock_hash):
        # given
        mock_hash.return_value = 'sha1', '123'

        # when
        with self.assertRaises(BadInputExc) as cm:
            query.parse_hash_with_algorithms_or_throws(
                'sha1:431',
                ['sha1_git'],
                'Only sha1_git!')
        self.assertIn('Only sha1_git!', cm.exception.args[0])

        mock_hash.assert_called_once_with('sha1:431')

    @patch('swh.web.common.query.parse_hash')
    def test_parse_hash_with_algorithms(self, mock_hash):
        # given
        mock_hash.return_value = ('sha256', b'123')

        # when
        algo, sha = query.parse_hash_with_algorithms_or_throws(
            'sha256:123',
            ['sha256', 'sha1_git'],
            'useless error message for this use case')

        self.assertEqual(algo, 'sha256')
        self.assertEqual(sha, b'123')

        mock_hash.assert_called_once_with('sha256:123')

    def test_parse_uuid4(self):
        # when
        actual_uuid = query.parse_uuid4('7c33636b-8f11-4bda-89d9-ba8b76a42cec')

        # then
        self.assertEqual(actual_uuid, '7c33636b-8f11-4bda-89d9-ba8b76a42cec')

    def test_parse_uuid4_ko(self):
        # when
        with self.assertRaises(BadInputExc) as cm:
            query.parse_uuid4('7c33636b-8f11-4bda-89d9-ba8b76a42')
        self.assertIn('badly formed hexadecimal UUID string',
                      cm.exception.args[0])



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from bs4 import BeautifulSoup
from htmlmin import minify


class HtmlPrettifyMiddleware(object):
    """
    Django middleware for prettifying generated HTML in
    development mode.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if 'text/html' in response.get('Content-Type', ''):
            response.content = \
                BeautifulSoup(response.content, 'lxml').prettify()

        return response


class HtmlMinifyMiddleware(object):
    """
    Django middleware for minifying generated HTML in
    production mode.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if 'text/html' in response.get('Content-Type', ''):
            try:
                minified_html = minify(response.content.decode('utf-8'),
                                       convert_charrefs=False)
                response.content = minified_html.encode('utf-8')
            except Exception:
                pass
        return response


class ThrottlingHeadersMiddleware(object):
    """
    Django middleware for inserting rate limiting related
    headers in HTTP response.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        resp = self.get_response(request)
        if 'RateLimit-Limit' in request.META:
            resp['X-RateLimit-Limit'] = request.META['RateLimit-Limit']
        if 'RateLimit-Remaining' in request.META:
            resp['X-RateLimit-Remaining'] = request.META['RateLimit-Remaining']
        if 'RateLimit-Reset' in request.META:
            resp['X-RateLimit-Reset'] = request.META['RateLimit-Reset']
        return resp



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.test import APITestCase
from unittest.mock import patch

from swh.web.common.exc import BadInputExc, NotFoundExc
from swh.web.common.utils import reverse
from swh.web.tests.testcase import WebTestCase

_snapshot_id = '36ce36946fcd0f39bdfc40727af4acfce81f67af'

_snapshot_branches = [
    {
        'name': 'gh-pages',
        'target': 'refs/heads/gh-pages',
        'target_type': 'alias',
    },
    {
        'name': 'latest',
        'target': 'refs/tags/v1.3.0',
        'target_type': 'alias',
    },
    {
        'name': 'refs/heads/andresgalante-dismissable-badges',
        'target': '3af57e6db525015a25b4f3abb29263432e4af7b1',
        'target_type': 'revision',
    },
    {
        'name': 'refs/heads/boom-toasted',
        'target': 'e3f4957facfbdc25fdc4c6d3f7fcf1c488f06cea',
        'target_type': 'revision',
    },
    {
        'name': 'refs/heads/flex-checks',
        'target': '1b08ea1630a987d6f56f0c99e869896485bf230b',
        'target_type': 'revision',
    },
    {
        'name': 'refs/heads/gh-pages',
        'target': 'ea4129886adec992483aa592db717f908b4be355',
        'target_type': 'revision',
    },
    {
        'name': 'refs/heads/grid-vertical-align-height',
        'target': 'ba23eea651f483b88b78bee1084d7d0f61161e8d',
        'target_type': 'revision',
    },
    {
        'name': 'refs/tags/v1.0.0',
        'target': 'd28343dc3ad53a411ae3685e7d6a7866c8c22d6b',
        'target_type': 'release',
    },
    {
        'name': 'refs/tags/v1.1.0',
        'target': '0f11410b77140852f835ad456e5fbcf28760846c',
        'target_type': 'release',
    },
    {
        'name': 'refs/tags/v1.1.1',
        'target': '4ca26e76ee4867bfcd65ecf81039f183fc1d3b4d',
        'target_type': 'release',
    },
    {
        'name': 'refs/tags/v1.2.0',
        'target': 'cabae13db21e0e1db686f5628d1725c6191062ef',
        'target_type': 'release',
    },
    {
        'name': 'refs/tags/v1.3.0',
        'target': '3c3d596d94501509bec1959a4cfb44b20bfa8606',
        'target_type': 'release',
    }
]


def _lookup_snapshot(snapshot_id, branches_from='',
                     branches_count=None, target_types=None):
    ret = {'id': snapshot_id, 'branches': {}}
    count = 0
    for branch in _snapshot_branches:
        if branches_count and count == branches_count:
            break
        if branch['name'] >= branches_from:
            if not target_types or branch['target_type'] in target_types: # noqa
                ret['branches'][branch['name']] = {
                    'target': branch['target'],
                    'target_type': branch['target_type']
                }
                count += 1
    return ret


def _get_branch_url(target_type, target):
    url = None
    if target_type == 'revision':
        url = reverse('api-revision', url_args={'sha1_git': target})
    if target_type == 'release':
        url = reverse('api-release', url_args={'sha1_git': target})
    return url


def _enrich_snapshot_data(snapshot_data):
    for branch in snapshot_data['branches'].keys():
        target = snapshot_data['branches'][branch]['target']
        target_type = snapshot_data['branches'][branch]['target_type']
        snapshot_data['branches'][branch]['target_url'] = \
            _get_branch_url(target_type, target)
    for branch in snapshot_data['branches'].keys():
        target = snapshot_data['branches'][branch]['target']
        target_type = snapshot_data['branches'][branch]['target_type']
        if target_type == 'alias':
            if target in snapshot_data['branches']:
                snapshot_data['branches'][branch]['target_url'] = \
                    snapshot_data['branches'][target]['target_url']
            else:
                snp = _lookup_snapshot(snapshot_data['id'],
                                       branches_from=target,
                                       branches_count=1)
                alias_target = snp['branches'][target]['target']
                alias_target_type = snp['branches'][target]['target_type']
                snapshot_data['branches'][branch]['target_url'] = \
                    _get_branch_url(alias_target_type, alias_target)

    return snapshot_data


@patch('swh.web.api.views.snapshot.service')
class SnapshotApiTestCase(WebTestCase, APITestCase):

    def test_api_snapshot(self, mock_service):
        mock_service.lookup_snapshot.side_effect = _lookup_snapshot

        url = reverse('api-snapshot',
                      url_args={'snapshot_id': _snapshot_id})
        rv = self.client.get(url)
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        expected_data = _lookup_snapshot(_snapshot_id)
        expected_data = _enrich_snapshot_data(expected_data)
        self.assertEqual(rv.data, expected_data)

    def test_api_snapshot_paginated(self, mock_service):
        mock_service.lookup_snapshot.side_effect = _lookup_snapshot

        branches_offset = 0
        branches_count = 2

        whole_snapshot = {'id': _snapshot_id, 'branches': {}}

        while branches_offset < len(_snapshot_branches):
            branches_from = _snapshot_branches[branches_offset]['name']
            url = reverse('api-snapshot',
                          url_args={'snapshot_id': _snapshot_id},
                          query_params={'branches_from': branches_from,
                                        'branches_count': branches_count})
            rv = self.client.get(url)
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv['Content-Type'], 'application/json')
            expected_data = _lookup_snapshot(_snapshot_id, branches_from,
                                             branches_count)
            expected_data = _enrich_snapshot_data(expected_data)
            self.assertEqual(rv.data, expected_data)
            whole_snapshot['branches'].update(expected_data['branches'])
            branches_offset += branches_count

            if branches_offset < len(_snapshot_branches):
                next_branch = _snapshot_branches[branches_offset]['name'] # noqa
                next_url = reverse('api-snapshot',
                                   url_args={'snapshot_id': _snapshot_id},
                                   query_params={'branches_from': next_branch,
                                                 'branches_count': branches_count}) # noqa
                self.assertEqual(rv['Link'], '<%s>; rel="next"' % next_url)
            else:
                self.assertFalse(rv.has_header('Link'))

        url = reverse('api-snapshot',
                      url_args={'snapshot_id': _snapshot_id})
        rv = self.client.get(url)
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, whole_snapshot)

    def test_api_snapshot_filtered(self, mock_service):
        mock_service.lookup_snapshot.side_effect = _lookup_snapshot

        target_types = 'release'

        url = reverse('api-snapshot',
                      url_args={'snapshot_id': _snapshot_id},
                      query_params={'target_types': target_types})
        rv = self.client.get(url)
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        expected_data = _lookup_snapshot(_snapshot_id,
                                         target_types=target_types.split(','))
        expected_data = _enrich_snapshot_data(expected_data)
        self.assertEqual(rv.data, expected_data)

    def test_api_snapshot_errors(self, mock_service):
        mock_service.lookup_snapshot.side_effect = \
            BadInputExc('Invalid snapshot id!')

        url = reverse('api-snapshot',
                      url_args={'snapshot_id': '63ce369'})
        rv = self.client.get(url)
        self.assertEqual(rv.status_code, 400)

        mock_service.lookup_snapshot.side_effect = \
            NotFoundExc('Snapshot not found!')

        snapshot_inexistent = '63ce36946fcd0f79bdfc40727af4acfce81f67fa'
        url = reverse('api-snapshot',
                      url_args={'snapshot_id': snapshot_inexistent})
        rv = self.client.get(url)
        self.assertEqual(rv.status_code, 404)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.conf.urls import url
from django.contrib.auth import views as auth_views
from django.shortcuts import redirect

from swh.web.admin.adminurls import AdminUrls

import swh.web.admin.origin_save # noqa
import swh.web.admin.deposit # noqa


def admin_default_view(request):
    return redirect('admin-origin-save')


urlpatterns = [url(r'^$', admin_default_view, name='admin'),
               url(r'^login/$', auth_views.login,
                   {'template_name': 'login.html'}, name='login'),
               url(r'^logout/$', auth_views.logout,
                   {'template_name': 'logout.html'}, name='logout')]

urlpatterns += AdminUrls.get_url_patterns()



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Utility module implementing Django views for browsing the archive
# in a snapshot context.
# Its purpose is to factorize code for the views reachable from the
# /origin/.* and /snapshot/.* endpoints.

from django.shortcuts import render, redirect
from django.template.defaultfilters import filesizeformat

from swh.model.identifiers import snapshot_identifier

from swh.web.browse.utils import (
    get_snapshot_context, get_directory_entries, gen_directory_link,
    gen_revision_link, request_content, gen_content_link,
    prepare_content_for_display, content_display_max_size,
    format_log_entries, gen_revision_log_link, gen_link,
    get_readme_to_display, get_swh_persistent_ids, process_snapshot_branches
)

from swh.web.common import service
from swh.web.common.exc import (
    handle_view_exception, NotFoundExc
)
from swh.web.common.utils import (
    reverse, gen_path_info, format_utc_iso_date, swh_object_icons
)

_empty_snapshot_id = snapshot_identifier({'branches': {}})


def _get_branch(branches, branch_name, snapshot_id):
    """
    Utility function to get a specific branch from a branches list.
    Its purpose is to get the default HEAD branch as some software origin
    (e.g those with svn type) does not have it. In that latter case, check
    if there is a master branch instead and returns it.
    """
    filtered_branches = \
        [b for b in branches if b['name'].endswith(branch_name)]
    if len(filtered_branches) > 0:
        return filtered_branches[0]
    elif branch_name == 'HEAD':
        filtered_branches = \
            [b for b in branches if b['name'].endswith('master')]
        if len(filtered_branches) > 0:
            return filtered_branches[0]
        elif len(branches) > 0:
            return branches[0]
    else:
        # case where a large branches list has been truncated
        snp = service.lookup_snapshot(snapshot_id,
                                      branches_from=branch_name,
                                      branches_count=1,
                                      target_types=['revision', 'alias'])
        snp_branch, _ = process_snapshot_branches(snp)
        if snp_branch:
            branches.append(snp_branch[0])
            return snp_branch[0]
    return None


def _get_release(releases, release_name):
    """
    Utility function to get a specific release from a releases list.
    Returns None if the release can not be found in the list.
    """
    filtered_releases = \
        [r for r in releases if r['name'] == release_name]
    if len(filtered_releases) > 0:
        return filtered_releases[0]
    else:
        return None


def _branch_not_found(branch_type, branch, branches, snapshot_id=None,
                      origin_info=None, timestamp=None, visit_id=None):
    """
    Utility function to raise an exception when a specified branch/release
    can not be found.
    """
    if branch_type == 'branch':
        branch_type = 'Branch'
        branch_type_plural = 'branches'
    else:
        branch_type = 'Release'
        branch_type_plural = 'releases'

    if snapshot_id and len(branches) == 0:
        msg = 'Snapshot with id %s has an empty list' \
              ' of %s!' % (snapshot_id, branch_type_plural)
    elif snapshot_id:
        msg = '%s %s for snapshot with id %s' \
              ' not found!' % (branch_type, branch, snapshot_id)
    elif visit_id and len(branches) == 0:
        msg = 'Origin with type %s and url %s' \
              ' for visit with id %s has an empty list' \
              ' of %s!' % (origin_info['type'], origin_info['url'], visit_id,
                           branch_type_plural)
    elif visit_id:
        msg = '%s %s associated to visit with' \
              ' id %s for origin with type %s and url %s' \
              ' not found!' % (branch_type, branch, visit_id,
                               origin_info['type'], origin_info['url'])
    elif len(branches) == 0:
        msg = 'Origin with type %s and url %s' \
                ' for visit with timestamp %s has an empty list' \
                ' of %s!' % (origin_info['type'], origin_info['url'],
                             timestamp, branch_type_plural)
    else:
        msg = '%s %s associated to visit with'  \
              ' timestamp %s for origin with type %s' \
              ' and url %s not found!' % (branch_type, branch, timestamp,
                                          origin_info['type'],
                                          origin_info['url'])
    raise NotFoundExc(msg)


def _process_snapshot_request(request, snapshot_id=None, origin_type=None,
                              origin_url=None, timestamp=None, path=None,
                              browse_context='directory'):
    """
    Utility function to perform common input request processing
    for snapshot context views.
    """

    visit_id = request.GET.get('visit_id', None)

    snapshot_context = get_snapshot_context(snapshot_id, origin_type,
                                            origin_url, timestamp, visit_id)

    swh_type = snapshot_context['swh_type']
    origin_info = snapshot_context['origin_info']
    branches = snapshot_context['branches']
    releases = snapshot_context['releases']
    url_args = snapshot_context['url_args']
    query_params = snapshot_context['query_params']

    if snapshot_context['visit_info']:
        timestamp = format_utc_iso_date(snapshot_context['visit_info']['date'],
                                        '%Y-%m-%dT%H:%M:%SZ')
        snapshot_context['timestamp'] = \
            format_utc_iso_date(snapshot_context['visit_info']['date'])

    browse_view_name = 'browse-' + swh_type + '-' + browse_context

    root_sha1_git = None
    revision_id = request.GET.get('revision', None)
    release_name = request.GET.get('release', None)
    release_id = None
    branch_name = None

    snapshot_total_size = sum(snapshot_context['snapshot_size'].values())

    if snapshot_total_size and revision_id:
        revision = service.lookup_revision(revision_id)
        root_sha1_git = revision['directory']
        branches.append({'name': revision_id,
                         'revision': revision_id,
                         'directory': root_sha1_git,
                         'url': None})
        branch_name = revision_id
        query_params['revision'] = revision_id
    elif snapshot_total_size and release_name:
        release = _get_release(releases, release_name)
        try:
            root_sha1_git = release['directory']
            revision_id = release['target']
            release_id = release['id']
            query_params['release'] = release_name
        except Exception:
            _branch_not_found("release", release_name, releases, snapshot_id,
                              origin_info, timestamp, visit_id)
    elif snapshot_total_size:
        branch_name = request.GET.get('branch', None)
        if branch_name:
            query_params['branch'] = branch_name
        branch = _get_branch(branches, branch_name or 'HEAD',
                             snapshot_context['snapshot_id'])
        try:
            branch_name = branch['name']
            revision_id = branch['revision']
            root_sha1_git = branch['directory']
        except Exception:
            _branch_not_found("branch", branch_name, branches, snapshot_id,
                              origin_info, timestamp, visit_id)

    for b in branches:
        branch_url_args = dict(url_args)
        branch_query_params = dict(query_params)
        if 'release' in branch_query_params:
            del branch_query_params['release']
        branch_query_params['branch'] = b['name']
        if path:
            b['path'] = path
            branch_url_args['path'] = path
        b['url'] = reverse(browse_view_name,
                           url_args=branch_url_args,
                           query_params=branch_query_params)

    for r in releases:
        release_url_args = dict(url_args)
        release_query_params = dict(query_params)
        if 'branch' in release_query_params:
            del release_query_params['branch']
        release_query_params['release'] = r['name']
        if path:
            r['path'] = path
            release_url_args['path'] = path
        r['url'] = reverse(browse_view_name,
                           url_args=release_url_args,
                           query_params=release_query_params)

    snapshot_context['query_params'] = query_params
    snapshot_context['root_sha1_git'] = root_sha1_git
    snapshot_context['revision_id'] = revision_id
    snapshot_context['branch'] = branch_name
    snapshot_context['release'] = release_name
    snapshot_context['release_id'] = release_id

    return snapshot_context


def browse_snapshot_directory(request, snapshot_id=None, origin_type=None,
                              origin_url=None, timestamp=None, path=None):
    """
    Django view implementation for browsing a directory in a snapshot context.
    """
    try:

        snapshot_context = _process_snapshot_request(request, snapshot_id,
                                                     origin_type, origin_url,
                                                     timestamp, path,
                                                     browse_context='directory') # noqa

        root_sha1_git = snapshot_context['root_sha1_git']
        sha1_git = root_sha1_git
        if root_sha1_git and path:
            dir_info = service.lookup_directory_with_path(root_sha1_git, path)
            # some readme files can reference assets reachable from the
            # browsed directory, handle that special case in order to
            # correctly displayed them
            if dir_info and dir_info['type'] == 'file':
                file_raw_url = reverse(
                    'browse-content-raw',
                    url_args={'query_string': dir_info['checksums']['sha1']})
                return redirect(file_raw_url)
            sha1_git = dir_info['target']

        dirs = []
        files = []
        if sha1_git:
            dirs, files = get_directory_entries(sha1_git)

    except Exception as exc:
        return handle_view_exception(request, exc)

    swh_type = snapshot_context['swh_type']
    origin_info = snapshot_context['origin_info']
    visit_info = snapshot_context['visit_info']
    url_args = snapshot_context['url_args']
    query_params = snapshot_context['query_params']
    revision_id = snapshot_context['revision_id']
    snapshot_id = snapshot_context['snapshot_id']

    path_info = gen_path_info(path)

    browse_view_name = 'browse-' + swh_type + '-directory'

    breadcrumbs = []
    if root_sha1_git:
        breadcrumbs.append({'name': root_sha1_git[:7],
                            'url': reverse(browse_view_name,
                                           url_args=url_args,
                                           query_params=query_params)})
    for pi in path_info:
        bc_url_args = dict(url_args)
        bc_url_args['path'] = pi['path']
        breadcrumbs.append({'name': pi['name'],
                            'url': reverse(browse_view_name,
                                           url_args=bc_url_args,
                                           query_params=query_params)})

    path = '' if path is None else (path + '/')

    for d in dirs:
        if d['type'] == 'rev':
            d['url'] = reverse('browse-revision',
                               url_args={'sha1_git': d['target']})
        else:
            bc_url_args = dict(url_args)
            bc_url_args['path'] = path + d['name']
            d['url'] = reverse(browse_view_name,
                               url_args=bc_url_args,
                               query_params=query_params)

    sum_file_sizes = 0

    readmes = {}

    browse_view_name = 'browse-' + swh_type + '-content'

    for f in files:
        bc_url_args = dict(url_args)
        bc_url_args['path'] = path + f['name']
        f['url'] = reverse(browse_view_name,
                           url_args=bc_url_args,
                           query_params=query_params)
        if f['length'] is not None:
            sum_file_sizes += f['length']
            f['length'] = filesizeformat(f['length'])
        if f['name'].lower().startswith('readme'):
            readmes[f['name']] = f['checksums']['sha1']

    readme_name, readme_url, readme_html = get_readme_to_display(readmes)

    browse_view_name = 'browse-' + swh_type + '-log'

    history_url = None
    if snapshot_id != _empty_snapshot_id:
        history_url = reverse(browse_view_name,
                              url_args=url_args,
                              query_params=query_params)

    nb_files = None
    nb_dirs = None
    sum_file_sizes = None
    dir_path = None
    if root_sha1_git:
        nb_files = len(files)
        nb_dirs = len(dirs)
        sum_file_sizes = filesizeformat(sum_file_sizes)
        dir_path = '/' + path

    browse_dir_link = \
        gen_directory_link(sha1_git, link_text='Browse',
                           link_attrs={'class': 'btn btn-default btn-sm',
                                       'role': 'button'})

    browse_rev_link = \
        gen_revision_link(revision_id,
                          snapshot_context=snapshot_context,
                          link_text='Browse',
                          link_attrs={'class': 'btn btn-default btn-sm',
                                      'role': 'button'})

    dir_metadata = {'id': sha1_git,
                    'context-independent directory': browse_dir_link,
                    'number of regular files': nb_files,
                    'number of subdirectories': nb_dirs,
                    'sum of regular file sizes': sum_file_sizes,
                    'path': dir_path,
                    'revision id': revision_id,
                    'revision': browse_rev_link,
                    'snapshot id': snapshot_id}

    if origin_info:
        dir_metadata['origin id'] = origin_info['id']
        dir_metadata['origin type'] = origin_info['type']
        dir_metadata['origin url'] = origin_info['url']
        dir_metadata['origin visit date'] = format_utc_iso_date(visit_info['date']) # noqa
        dir_metadata['origin visit id'] = visit_info['visit']
        snapshot_context_url = reverse('browse-snapshot-directory',
                                       url_args={'snapshot_id': snapshot_id},
                                       query_params=request.GET)
        browse_snapshot_link = \
            gen_link(snapshot_context_url, link_text='Browse',
                     link_attrs={'class': 'btn btn-default btn-sm',
                                 'role': 'button'})
        dir_metadata['snapshot context'] = browse_snapshot_link

    vault_cooking = {
        'directory_context': True,
        'directory_id': sha1_git,
        'revision_context': True,
        'revision_id': revision_id
    }

    swh_objects = [{'type': 'directory',
                    'id': sha1_git},
                   {'type': 'revision',
                    'id': revision_id},
                   {'type': 'snapshot',
                    'id': snapshot_id}]

    release_id = snapshot_context['release_id']
    if release_id:
        swh_objects.append({'type': 'release',
                            'id': release_id})

    swh_ids = get_swh_persistent_ids(swh_objects, snapshot_context)

    dir_path = '/'.join([bc['name'] for bc in breadcrumbs]) + '/'
    context_found = 'snapshot: %s' % snapshot_context['snapshot_id']
    if origin_info:
        context_found = 'origin: %s' % origin_info['url']
    heading = 'Directory - %s - %s - %s' %\
        (dir_path, snapshot_context['branch'], context_found)

    return render(request, 'browse/directory.html',
                  {'heading': heading,
                   'swh_object_name': 'Directory',
                   'swh_object_metadata': dir_metadata,
                   'dirs': dirs,
                   'files': files,
                   'breadcrumbs': breadcrumbs if root_sha1_git else [],
                   'top_right_link': {
                       'url': history_url,
                       'icon': swh_object_icons['revisions history'],
                       'text': 'History'
                    },
                   'readme_name': readme_name,
                   'readme_url': readme_url,
                   'readme_html': readme_html,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': vault_cooking,
                   'show_actions_menu': True,
                   'swh_ids': swh_ids})


def browse_snapshot_content(request, snapshot_id=None, origin_type=None,
                            origin_url=None, timestamp=None, path=None):
    """
    Django view implementation for browsing a content in a snapshot context.
    """
    try:

        snapshot_context = _process_snapshot_request(request, snapshot_id,
                                                     origin_type, origin_url,
                                                     timestamp, path,
                                                     browse_context='content')

        root_sha1_git = snapshot_context['root_sha1_git']
        sha1_git = None
        query_string = None
        content_data = None
        if root_sha1_git:
            content_info = service.lookup_directory_with_path(root_sha1_git,
                                                              path)
            sha1_git = content_info['target']
            query_string = 'sha1_git:' + sha1_git
            content_data = request_content(query_string,
                                           raise_if_unavailable=False)

    except Exception as exc:
        return handle_view_exception(request, exc)

    swh_type = snapshot_context['swh_type']
    url_args = snapshot_context['url_args']
    query_params = snapshot_context['query_params']
    revision_id = snapshot_context['revision_id']
    origin_info = snapshot_context['origin_info']
    visit_info = snapshot_context['visit_info']
    snapshot_id = snapshot_context['snapshot_id']

    content = None
    language = None
    mimetype = None
    if content_data and content_data['raw_data'] is not None:
        content_display_data = prepare_content_for_display(
            content_data['raw_data'], content_data['mimetype'], path)
        content = content_display_data['content_data']
        language = content_display_data['language']
        mimetype = content_display_data['mimetype']

    filename = None
    path_info = None

    browse_view_name = 'browse-' + swh_type + '-directory'

    breadcrumbs = []

    split_path = path.split('/')
    filename = split_path[-1]
    path_info = gen_path_info(path[:-len(filename)])
    if root_sha1_git:
        breadcrumbs.append({'name': root_sha1_git[:7],
                            'url': reverse(browse_view_name,
                                           url_args=url_args,
                                           query_params=query_params)})
    for pi in path_info:
        bc_url_args = dict(url_args)
        bc_url_args['path'] = pi['path']
        breadcrumbs.append({'name': pi['name'],
                            'url': reverse(browse_view_name,
                                           url_args=bc_url_args,
                                           query_params=query_params)})

    breadcrumbs.append({'name': filename,
                        'url': None})

    browse_content_link = \
        gen_content_link(sha1_git, link_text='Browse',
                         link_attrs={'class': 'btn btn-default btn-sm',
                                     'role': 'button'})

    content_raw_url = None
    if query_string:
        content_raw_url = reverse('browse-content-raw',
                                  url_args={'query_string': query_string},
                                  query_params={'filename': filename})

    browse_rev_link = \
        gen_revision_link(revision_id,
                          snapshot_context=snapshot_context,
                          link_text='Browse',
                          link_attrs={'class': 'btn btn-default btn-sm',
                                      'role': 'button'})

    content_metadata = {
        'context-independent content': browse_content_link,
        'path': None,
        'filename': None,
        'revision id': revision_id,
        'revision': browse_rev_link,
        'snapshot id': snapshot_id
    }

    cnt_sha1_git = None
    content_size = None
    error_code = 200
    error_description = ''
    error_message = ''
    if content_data:
        content_metadata['sha1 checksum'] = \
            content_data['checksums']['sha1']
        content_metadata['sha1_git checksum'] = \
            content_data['checksums']['sha1_git']
        content_metadata['sha256 checksum'] = \
            content_data['checksums']['sha256']
        content_metadata['blake2s256 checksum'] = \
            content_data['checksums']['blake2s256']
        content_metadata['mime type'] = content_data['mimetype']
        content_metadata['encoding'] = content_data['encoding']
        content_metadata['size'] = filesizeformat(content_data['length'])
        content_metadata['language'] = content_data['language']
        content_metadata['licenses'] = content_data['licenses']
        content_metadata['path'] = '/' + path[:-len(filename)]
        content_metadata['filename'] = filename

        cnt_sha1_git = content_data['checksums']['sha1_git']
        content_size = content_data['length']
        error_code = content_data['error_code']
        error_message = content_data['error_message']
        error_description = content_data['error_description']

    if origin_info:
        content_metadata['origin id'] = origin_info['id']
        content_metadata['origin type'] = origin_info['type']
        content_metadata['origin url'] = origin_info['url']
        content_metadata['origin visit date'] = format_utc_iso_date(visit_info['date']) # noqa
        content_metadata['origin visit id'] = visit_info['visit']
        browse_snapshot_url = reverse('browse-snapshot-content',
                                      url_args={'snapshot_id': snapshot_id,
                                                'path': path},
                                      query_params=request.GET)
        browse_snapshot_link = \
            gen_link(browse_snapshot_url, link_text='Browse',
                     link_attrs={'class': 'btn btn-default btn-sm',
                                 'role': 'button'})
        content_metadata['snapshot context'] = browse_snapshot_link

    swh_objects = [{'type': 'content',
                    'id': cnt_sha1_git},
                   {'type': 'revision',
                    'id': revision_id},
                   {'type': 'snapshot',
                    'id': snapshot_id}]

    release_id = snapshot_context['release_id']
    if release_id:
        swh_objects.append({'type': 'release',
                            'id': release_id})

    swh_ids = get_swh_persistent_ids(swh_objects, snapshot_context)

    content_path = '/'.join([bc['name'] for bc in breadcrumbs])
    context_found = 'snapshot: %s' % snapshot_context['snapshot_id']
    if origin_info:
        context_found = 'origin: %s' % origin_info['url']
    heading = 'Content - %s - %s - %s' %\
        (content_path, snapshot_context['branch'], context_found)

    return render(request, 'browse/content.html',
                  {'heading': heading,
                   'swh_object_name': 'Content',
                   'swh_object_metadata': content_metadata,
                   'content': content,
                   'content_size': content_size,
                   'max_content_size': content_display_max_size,
                   'mimetype': mimetype,
                   'language': language,
                   'breadcrumbs': breadcrumbs if root_sha1_git else [],
                   'top_right_link': {
                        'url': content_raw_url,
                        'icon': swh_object_icons['content'],
                        'text': 'Raw File'
                    },
                   'snapshot_context': snapshot_context,
                   'vault_cooking': None,
                   'show_actions_menu': True,
                   'swh_ids': swh_ids,
                   'error_code': error_code,
                   'error_message': error_message,
                   'error_description': error_description},
                  status=error_code)


PER_PAGE = 100


def browse_snapshot_log(request, snapshot_id=None, origin_type=None,
                        origin_url=None, timestamp=None):
    """
    Django view implementation for browsing a revision history in a
    snapshot context.
    """
    try:

        snapshot_context = _process_snapshot_request(request, snapshot_id,
                                                     origin_type, origin_url,
                                                     timestamp, browse_context='log') # noqa

        revision_id = snapshot_context['revision_id']

        per_page = int(request.GET.get('per_page', PER_PAGE))
        offset = int(request.GET.get('offset', 0))
        revs_ordering = request.GET.get('revs_ordering', 'committer_date')
        session_key = 'rev_%s_log_ordering_%s' % (revision_id, revs_ordering)
        rev_log_session = request.session.get(session_key, None)
        rev_log = []
        revs_walker_state = None
        if rev_log_session:
            rev_log = rev_log_session['rev_log']
            revs_walker_state = rev_log_session['revs_walker_state']

        if len(rev_log) < offset+per_page:
            revs_walker = \
                service.get_revisions_walker(revs_ordering,
                                             revision_id,
                                             max_revs=offset+per_page+1,
                                             state=revs_walker_state)
            rev_log += list(revs_walker)
            revs_walker_state = revs_walker.export_state()

        revision_log = rev_log[offset:offset+per_page]

        request.session[session_key] = {
            'rev_log': rev_log,
            'revs_walker_state': revs_walker_state
        }

    except Exception as exc:
        return handle_view_exception(request, exc)

    swh_type = snapshot_context['swh_type']
    origin_info = snapshot_context['origin_info']
    visit_info = snapshot_context['visit_info']
    url_args = snapshot_context['url_args']
    query_params = snapshot_context['query_params']
    snapshot_id = snapshot_context['snapshot_id']

    query_params['per_page'] = per_page
    revs_ordering = request.GET.get('revs_ordering', '')
    query_params['revs_ordering'] = revs_ordering

    browse_view_name = 'browse-' + swh_type + '-log'

    prev_log_url = None
    if len(rev_log) > offset + per_page:
        query_params['offset'] = offset + per_page
        prev_log_url = reverse(browse_view_name,
                               url_args=url_args,
                               query_params=query_params)

    next_log_url = None
    if offset != 0:
        query_params['offset'] = offset - per_page
        next_log_url = reverse(browse_view_name,
                               url_args=url_args,
                               query_params=query_params)

    revision_log_data = format_log_entries(revision_log, per_page,
                                           snapshot_context)

    browse_log_link = \
        gen_revision_log_link(revision_id, link_text='Browse',
                              link_attrs={'class': 'btn btn-default btn-sm',
                                          'role': 'button'})

    revision_metadata = {
        'context-independent revision history': browse_log_link,
        'snapshot id': snapshot_id
    }

    if origin_info:
        revision_metadata['origin id'] = origin_info['id']
        revision_metadata['origin type'] = origin_info['type']
        revision_metadata['origin url'] = origin_info['url']
        revision_metadata['origin visit date'] = format_utc_iso_date(visit_info['date']) # noqa
        revision_metadata['origin visit id'] = visit_info['visit']
        browse_snapshot_url = reverse('browse-snapshot-log',
                                      url_args={'snapshot_id': snapshot_id},
                                      query_params=request.GET)
        browse_snapshot_link = \
            gen_link(browse_snapshot_url, link_text='Browse',
                     link_attrs={'class': 'btn btn-default btn-sm',
                                 'role': 'button'})
        revision_metadata['snapshot context'] = browse_snapshot_link

    swh_objects = [{'type': 'revision',
                    'id': revision_id},
                   {'type': 'snapshot',
                    'id': snapshot_id}]

    release_id = snapshot_context['release_id']
    if release_id:
        swh_objects.append({'type': 'release',
                            'id': release_id})

    swh_ids = get_swh_persistent_ids(swh_objects, snapshot_context)

    context_found = 'snapshot: %s' % snapshot_context['snapshot_id']
    if origin_info:
        context_found = 'origin: %s' % origin_info['url']
    heading = 'Revision history - %s - %s' %\
        (snapshot_context['branch'], context_found)

    return render(request, 'browse/revision-log.html',
                  {'heading': heading,
                   'swh_object_name': 'Revisions history',
                   'swh_object_metadata': revision_metadata,
                   'revision_log': revision_log_data,
                   'revs_ordering': revs_ordering,
                   'next_log_url': next_log_url,
                   'prev_log_url': prev_log_url,
                   'breadcrumbs': None,
                   'top_right_link': None,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': None,
                   'show_actions_menu': True,
                   'swh_ids': swh_ids})


def browse_snapshot_branches(request, snapshot_id=None, origin_type=None,
                             origin_url=None, timestamp=None):
    """
    Django view implementation for browsing a list of branches in a snapshot
    context.
    """
    try:
        snapshot_context = _process_snapshot_request(request, snapshot_id,
                                                     origin_type, origin_url,
                                                     timestamp)

        branches_bc = request.GET.get('branches_breadcrumbs', '')
        branches_bc = \
            branches_bc.split(',') if branches_bc else []
        branches_from = branches_bc[-1] if branches_bc else ''

        swh_type = snapshot_context['swh_type']
        origin_info = snapshot_context['origin_info']
        url_args = snapshot_context['url_args']
        query_params = snapshot_context['query_params']

        browse_view_name = 'browse-' + swh_type + '-directory'

        snapshot = \
            service.lookup_snapshot(snapshot_context['snapshot_id'],
                                    branches_from, PER_PAGE+1,
                                    target_types=['revision', 'alias'])

        displayed_branches, _ = process_snapshot_branches(snapshot)

    except Exception as exc:
        return handle_view_exception(request, exc)

    for branch in displayed_branches:
        if snapshot_id:
            revision_url = reverse('browse-revision',
                                   url_args={'sha1_git': branch['revision']},
                                   query_params={'snapshot_id': snapshot_id})
        else:
            revision_url = reverse('browse-revision',
                                   url_args={'sha1_git': branch['revision']},
                                   query_params={'origin_type': origin_type,
                                                 'origin': origin_info['url']})
        query_params['branch'] = branch['name']
        directory_url = reverse(browse_view_name,
                                url_args=url_args,
                                query_params=query_params)
        del query_params['branch']
        branch['revision_url'] = revision_url
        branch['directory_url'] = directory_url

    browse_view_name = 'browse-' + swh_type + '-branches'

    prev_branches_url = None
    next_branches_url = None

    if branches_bc:
        query_params_prev = dict(query_params)

        query_params_prev['branches_breadcrumbs'] = \
            ','.join(branches_bc[:-1])
        prev_branches_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params_prev)
    elif branches_from:
        prev_branches_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params)

    if len(displayed_branches) > PER_PAGE:
        query_params_next = dict(query_params)
        next_branch = displayed_branches[-1]['name']
        del displayed_branches[-1]
        branches_bc.append(next_branch)
        query_params_next['branches_breadcrumbs'] = \
            ','.join(branches_bc)
        next_branches_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params_next)

    heading = 'Branches - '
    if origin_info:
        heading += 'origin: %s' % origin_info['url']
    else:
        heading += 'snapshot: %s' % snapshot_id

    return render(request, 'browse/branches.html',
                  {'heading': heading,
                   'swh_object_name': 'Branches',
                   'swh_object_metadata': {},
                   'top_right_link': None,
                   'displayed_branches': displayed_branches,
                   'prev_branches_url': prev_branches_url,
                   'next_branches_url': next_branches_url,
                   'snapshot_context': snapshot_context})


def browse_snapshot_releases(request, snapshot_id=None, origin_type=None,
                             origin_url=None, timestamp=None):
    """
    Django view implementation for browsing a list of releases in a snapshot
    context.
    """
    try:
        snapshot_context = _process_snapshot_request(request, snapshot_id,
                                                     origin_type, origin_url,
                                                     timestamp)

        rel_bc = request.GET.get('releases_breadcrumbs', '')
        rel_bc = \
            rel_bc.split(',') if rel_bc else []
        rel_from = rel_bc[-1] if rel_bc else ''

        swh_type = snapshot_context['swh_type']
        origin_info = snapshot_context['origin_info']
        url_args = snapshot_context['url_args']
        query_params = snapshot_context['query_params']

        snapshot = \
            service.lookup_snapshot(snapshot_context['snapshot_id'],
                                    rel_from, PER_PAGE+1,
                                    target_types=['release', 'alias'])

        _, displayed_releases = process_snapshot_branches(snapshot)

    except Exception as exc:
        return handle_view_exception(request, exc)

    for release in displayed_releases:
        if snapshot_id:
            query_params_tgt = {'snapshot_id': snapshot_id}
        else:
            query_params_tgt = {'origin': origin_info['url']}
        release_url = reverse('browse-release',
                              url_args={'sha1_git': release['id']},
                              query_params=query_params_tgt)

        target_url = ''
        if release['target_type'] == 'revision':
            target_url = reverse('browse-revision',
                                 url_args={'sha1_git': release['target']},
                                 query_params=query_params_tgt)
        elif release['target_type'] == 'directory':
            target_url = reverse('browse-directory',
                                 url_args={'sha1_git': release['target']},
                                 query_params=query_params_tgt)
        elif release['target_type'] == 'content':
            target_url = reverse('browse-content',
                                 url_args={'query_string': release['target']},
                                 query_params=query_params_tgt)
        elif release['target_type'] == 'release':
            target_url = reverse('browse-release',
                                 url_args={'sha1_git': release['target']},
                                 query_params=query_params_tgt)

        release['release_url'] = release_url
        release['target_url'] = target_url

    browse_view_name = 'browse-' + swh_type + '-releases'

    prev_releases_url = None
    next_releases_url = None

    if rel_bc:
        query_params_prev = dict(query_params)

        query_params_prev['releases_breadcrumbs'] = \
            ','.join(rel_bc[:-1])
        prev_releases_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params_prev)
    elif rel_from:
        prev_releases_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params)

    if len(displayed_releases) > PER_PAGE:
        query_params_next = dict(query_params)
        next_rel = displayed_releases[-1]['branch_name']
        del displayed_releases[-1]
        rel_bc.append(next_rel)
        query_params_next['releases_breadcrumbs'] = \
            ','.join(rel_bc)
        next_releases_url = reverse(browse_view_name, url_args=url_args,
                                    query_params=query_params_next)

    heading = 'Releases - '
    if origin_info:
        heading += 'origin: %s' % origin_info['url']
    else:
        heading += 'snapshot: %s' % snapshot_id

    return render(request, 'browse/releases.html',
                  {'heading': heading,
                   'top_panel_visible': False,
                   'top_panel_collapsible': False,
                   'swh_object_name': 'Releases',
                   'swh_object_metadata': {},
                   'top_right_link': None,
                   'displayed_releases': displayed_releases,
                   'prev_releases_url': prev_releases_url,
                   'next_releases_url': next_releases_url,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': None,
                   'show_actions_menu': False})
